#! /usr/bin/env python3
# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

import os
import sys
import shutil
import argparse
from os.path import join, abspath, dirname

sys.path.insert(0, dirname(dirname(abspath(__file__))))
from pl_setup.packaging_base import (
    Host,
    SB_Dir,
    DataTree,
    PDFium_URL,
    DepotTools_URL,
    MainLibnames,
    LibnameForSystem,
    VerStatusFileName,
    PlatformNames,
    run_cmd,
    call_ctypesgen,
)


PatchDir       = join(SB_Dir, "patches")
DepotToolsDir  = join(SB_Dir, "depot_tools")
PDFiumDir      = join(SB_Dir, "pdfium")
PDFiumBuildDir = join(PDFiumDir, "out", "Default")
OutputDir      = join(DataTree, PlatformNames.sourcebuild)

PdfiumMainPatches = [
    (join(PatchDir, "public_headers.patch"), PDFiumDir),
    (join(PatchDir, "shared_library.patch"), PDFiumDir),
]
PdfiumWinPatches = [
    (join(PatchDir, "win", "pdfium.patch"), PDFiumDir),
    (join(PatchDir, "win", "build.patch"), join(PDFiumDir, "build")),
]

DefaultConfig = {
    "is_debug": False,
    "treat_warnings_as_errors": False,
    # "clang_use_chrome_plugins": False,
    "pdf_is_standalone": True,
    "pdf_enable_v8": False,
    "pdf_enable_xfa": False,
    "pdf_use_skia": False,
}

if sys.platform.startswith("darwin"):
    DefaultConfig["mac_deployment_target"] = "10.11.0"
elif sys.platform.startswith("win32"):
    DefaultConfig["pdf_use_win32_gdi"] = True


def dl_depottools(do_update):
    
    if not os.path.isdir(SB_Dir):
        os.makedirs(SB_Dir)
    
    is_update = True
    
    if os.path.isdir(DepotToolsDir):
        if do_update:
            print("DepotTools: Revert and update ...")
            run_cmd(["git", "reset", "--hard", "HEAD"], cwd=DepotToolsDir)
            run_cmd(["git", "pull", DepotTools_URL], cwd=DepotToolsDir)
        else:
            print("DepotTools: Using existing repository as-is.")
            is_update = False
    else:
        print("DepotTools: Download ...")
        run_cmd(["git", "clone", "--depth", "1", DepotTools_URL, DepotToolsDir], cwd=SB_Dir)
    
    os.environ["PATH"] += os.pathsep + DepotToolsDir
    
    return is_update


def dl_pdfium(GClient, do_update, revision):
    
    is_sync = True
    
    if os.path.isdir(PDFiumDir):
        if do_update:
            print("PDFium: Revert / Sync  ...")
            run_cmd([GClient, "revert"], cwd=SB_Dir)
        else:
            is_sync = False
            print("PDFium: Using existing repository as-is.")
    else:
        print("PDFium: Download ...")
        run_cmd([GClient, "config", "--custom-var", "checkout_configuration=minimal", "--unmanaged", PDFium_URL], cwd=SB_Dir)
    
    # TODO consider --with_branch_heads so we can improve get_pdfium_version()
    # while the heads take up additional disk space, this shouldn't be a problem for a shallow clone
    if is_sync:
        run_cmd([GClient, "sync", "--revision", "origin/%s" % revision, "--no-history", "--shallow"], cwd=SB_Dir)
    
    return is_sync


def get_pdfium_version():
    
    # FIXME awkward mix of local/remote git - this will fail to identify the tag if local and remote state do not match
    
    head_commit = run_cmd(["git", "rev-parse", "--short", "HEAD"], cwd=PDFiumDir, capture=True)
    refs_string = run_cmd(["git", "ls-remote", "--heads", PDFium_URL, "chromium/*"], cwd=None, capture=True)
    
    latest = refs_string.split("\n")[-1]
    tag_commit, ref = latest.split("\t")
    tag_commit = tag_commit[:7]
    tag = ref.split("/")[-1]
    
    print("Current head %s, latest tagged commit %s (%s)" % (head_commit, tag_commit, tag))
    
    if head_commit == tag_commit:
        v_libpdfium = tag
    else:
        v_libpdfium = head_commit
    
    return v_libpdfium


def update_version(v_libpdfium):
    ver_file = join(OutputDir, VerStatusFileName)
    with open(ver_file, "w") as fh:
        fh.write( str(v_libpdfium) )


def _apply_patchset(patchset):
    for patch, cwd in patchset:
        run_cmd(["git", "apply", "-v", patch], cwd=cwd)


def _create_resources_rc(v_libpdfium):
    
    input_path = join(PatchDir, "win", "resources.rc")
    output_path = join(PDFiumDir, "resources.rc")
    
    with open(input_path, "r") as fh:
        content = fh.read()
    
    vars = ("$VERSION_CSV", "$VERSION")
    for var in vars:
        content = content.replace(var, v_libpdfium)
    
    with open(output_path, "w") as fh:
        fh.write(content)


def patch_pdfium(v_libpdfium):
    _apply_patchset(PdfiumMainPatches)
    if sys.platform.startswith("win32"):
        _apply_patchset(PdfiumWinPatches)
        _create_resources_rc(v_libpdfium)


def configure(GN, config):
    if not os.path.exists(PDFiumBuildDir):
        os.makedirs(PDFiumBuildDir)
    with open(join(PDFiumBuildDir, "args.gn"), "w") as args_handle:
        args_handle.write(config)
    run_cmd([GN, "gen", PDFiumBuildDir], cwd=PDFiumDir)


def build(Ninja, target):
    run_cmd([Ninja, "-C", PDFiumBuildDir, target], cwd=PDFiumDir)


def find_lib(src_libname=None, directory=PDFiumBuildDir):
    
    if src_libname is not None:
        path = join(PDFiumBuildDir, src_libname)
        if os.path.isfile(path):
            return path
        else:
            print("Warning: Binary not found under given name.", file=sys.stderr)
    
    try_names = []
    for name in MainLibnames + ["pdfium.so"]:
        try_names += [name, "lib"+name]
    
    try_paths = [join(directory, n) for n in try_names]
    found_paths = [fp for fp in try_paths if os.path.isfile(fp)]
    
    assert len(found_paths) == 1
    return found_paths[0]


def pack(src_libpath, v_libpdfium, destname=None):
    
    # TODO remove existing binary/bindings, just to be safe
    
    if destname is None:
        destname = LibnameForSystem[Host.system]
    
    destpath = join(OutputDir, destname)
    shutil.copy(src_libpath, destpath)
    
    update_version(v_libpdfium)
    
    include_dir = join(PDFiumDir, "public")
    call_ctypesgen(OutputDir, include_dir)


def get_tool(tool, win_append):
    exe = join(DepotToolsDir, tool)
    if sys.platform.startswith("win32"):
        exe += "." + win_append
    return exe


def serialise_config(config_dict):
    
    config_str = ""
    sep = ""
    
    for key, value in config_dict.items():
        config_str += sep + "%s = " % key
        if isinstance(value, bool):
            config_str += str(value).lower()
        elif isinstance(value, str):
            config_str += '"%s"' % value
        else:
            raise TypeError("Not sure how to serialise type %s" % type(value))
        sep = "\n"
    
    return config_str


def main(
        b_src_libname = None,
        b_dest_libname = None,
        b_update = False,
        b_revision = None,
        b_target = None,
    ):
    
    if not os.path.exists(OutputDir):
        os.makedirs(OutputDir)
    
    if b_revision is None:
        b_revision = "main"
    if b_target is None:
        b_target = "pdfium"
    
    if sys.platform.startswith("win32"):
        os.environ["DEPOT_TOOLS_WIN_TOOLCHAIN"] = "0"
    
    dl_depottools(b_update)
    
    GClient = get_tool("gclient", "bat")
    GN      = get_tool("gn", "bat")
    Ninja   = get_tool("ninja", "exe")
    
    pdfium_dl_done = dl_pdfium(GClient, b_update, b_revision)
    v_libpdfium = get_pdfium_version()
    if pdfium_dl_done:
        patch_pdfium(v_libpdfium)
    
    config_dict = DefaultConfig.copy()
    config_str = serialise_config(config_dict)
    print("\nBuild configuration:\n%s\n" % config_str)
    
    configure(GN, config_str)
    build(Ninja, b_target)
    libpath = find_lib(b_src_libname)
    pack(libpath, v_libpdfium, b_dest_libname)


def parse_args(argv):
    
    parser = argparse.ArgumentParser(
        description = "A script to automate building PDFium from source and generating bindings with ctypesgen.",
    )
    
    parser.add_argument(
        "--src-libname",
        help = "Name of the generated PDFium binary file. This script tries to automatically find the binary, which should usually work. If it does not, however, this option may be used to explicitly provide the file name to look for.",
    )
    parser.add_argument(
        "--dest-libname",
        help = "Rename the binary. Must be a name recognised by packaging code.",
    )
    parser.add_argument(
        "--update", "-u",
        action = "store_true",
        help = "Update existing PDFium/DepotTools repositories, removing local changes.",
    )
    parser.add_argument(
        "--revision", "-r",
        help = "PDFium revision to check out (defaults to main).",
    )
    parser.add_argument(
        "--target", "-t",
        help = "PDFium build target (defaults to `pdfium`). Use `pdfium_all` to also build tests."
    )
    
    return parser.parse_args(argv)


def main_cli(argv=sys.argv[1:]):
    args = parse_args(argv)
    return main(
        b_src_libname  = args.src_libname,
        b_dest_libname = args.dest_libname,
        b_update   = args.update,
        b_revision = args.revision,
        b_target   = args.target,
    )
    

if __name__ == "__main__":
    main_cli()



#! /usr/bin/env python3
# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

import os
import sys
import shutil
import tarfile
import argparse
import traceback
import functools
from urllib import request
from os.path import join, abspath, dirname
from concurrent.futures import ThreadPoolExecutor

sys.path.insert(0, dirname(dirname(abspath(__file__))))
from pl_setup.packaging_base import (
    Host,
    DataTree,
    VerStatusFileName,
    ReleaseNames,
    BinaryPlatforms,
    ReleaseURL,
    BinaryTarget_Auto,
    get_latest_version,
    call_ctypesgen,
)


def clear_data(download_files):
    for pl_name in download_files:
        pl_dir = join(DataTree, pl_name)
        if os.path.isdir(pl_dir):
            shutil.rmtree(pl_dir)


def _get_package(latest_ver, robust, pl_name):
    
    pl_dir = join(DataTree, pl_name)
    if not os.path.exists(pl_dir):
        os.makedirs(pl_dir)
    
    file_name = "%s.%s" % (ReleaseNames[pl_name], "tgz")
    file_url = "%s%s/%s" % (ReleaseURL, latest_ver, file_name)
    file_path = join(pl_dir, file_name)
    print("'%s' -> '%s'" % (file_url, file_path))
    
    try:
        request.urlretrieve(file_url, file_path)
    except Exception:
        if robust:
            traceback.print_exc()
            return None, None
        else:
            raise
    
    return pl_name, file_path


def download_releases(latest_ver, platforms, robust, max_workers):
    if not max_workers:
        max_workers = len(platforms)
    archives = {}
    with ThreadPoolExecutor(max_workers=max_workers) as pool:
        func = functools.partial(_get_package, latest_ver, robust)
        for pl_name, file_path in pool.map(func, platforms):
            if pl_name is None:
                continue
            archives[pl_name] = file_path
    return archives


def unpack_archives(archives):
    for pl_name, file_path in archives.items():
        extraction_path = join(DataTree, pl_name, "build_tar")
        with tarfile.open(file_path) as archive:
            archive.extractall(extraction_path)
        os.remove(file_path)


def generate_bindings(archives, latest_ver):
    
    for pl_name in archives.keys():
        
        pl_dir = join(DataTree, pl_name)
        build_dir = join(pl_dir, "build_tar")
        bin_dir = join(build_dir, "lib")
        dirname = os.path.basename(pl_dir)
        
        if dirname.startswith("windows"):
            target_name = "pdfium.dll"
            bin_dir = join(build_dir, "bin")
        elif dirname.startswith("darwin"):
            target_name = "pdfium.dylib"
        elif "linux" in dirname:
            target_name = "pdfium"
        else:
            raise ValueError("Unknown platform directory name '%s'" % dirname)
        
        items = os.listdir(bin_dir)
        assert len(items) == 1
        shutil.move(join(bin_dir, items[0]), join(pl_dir, target_name))
        
        ver_file = join(DataTree, pl_name, VerStatusFileName)
        with open(ver_file, "w") as fh:
            fh.write(latest_ver)
        
        call_ctypesgen(pl_dir, join(build_dir, "include"))
        shutil.rmtree(build_dir)


def main(platforms, robust=False, max_workers=None):
    
    if len(platforms) != len(set(platforms)):
        raise ValueError("Duplicate platforms not allowed.")
    if BinaryTarget_Auto in platforms:
        platforms = platforms.copy()
        platforms[platforms.index(BinaryTarget_Auto)] = Host.platform
    
    latest_ver = str( get_latest_version() )
    clear_data(platforms)
    
    archives = download_releases(latest_ver, platforms, robust, max_workers)
    unpack_archives(archives)
    generate_bindings(archives, latest_ver)


def parse_args(argv):
    platform_choices = (BinaryTarget_Auto, *BinaryPlatforms)
    parser = argparse.ArgumentParser(
        description = "Download pre-built PDFium packages and generate bindings.",
    )
    parser.add_argument(
        "--platforms", "-p",
        nargs = "+",
        metavar = "identifier",
        choices = platform_choices,
        default = BinaryPlatforms,
        help = "The platform(s) to include. Available platform identifiers are %s. `auto` represents the current host platform." % (platform_choices, ),
    )
    parser.add_argument(
        "--robust",
        action = "store_true",
        help = "Skip missing binaries instead of raising an exception.",
    )
    parser.add_argument(
        "--max-workers",
        type = int,
        help = "Maximum number of jobs to run in parallel when downloading binaries.",
    )
    return parser.parse_args(argv)


def run_cli(argv=sys.argv[1:]):
    args = parse_args(argv)
    main(
        args.platforms,
        robust = args.robust,
        max_workers = args.max_workers,
    )


if __name__ == "__main__":
    run_cli()



# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

import os
from os.path import (
    join,
    abspath,
    basename,
    splitext,
)
from pypdfium2 import _namespace as pdfium
from pypdfium2._cli._parsers import pagetext_type


ColourOpts = dict(
    metavar = "C",
    nargs = 4,
    type = int,
)


def attach_parser(subparsers):
    parser = subparsers.add_parser(
        "render",
        help = "Rasterise pages of a PDF file",
    )
    parser.add_argument(
        "inputs",
        nargs = "+",
        help = "PDF documents to render",
    )
    parser.add_argument(
        "--passwords",
        nargs = "*",
        help = "A sequence of passwords to unlock encrypted PDFs. The value is ignored for non-encrypted documents, where any placeholder may be used.",
    )
    parser.add_argument(
        "--output", "-o",
        type = abspath,
        required = True,
        help = "Output directory where to place the serially numbered images",
    )
    parser.add_argument(
        "--format", "-f",
        default = "jpg",
        help = "The image format to use",
    )
    parser.add_argument(
        "--pages",
        default = None,
        type = pagetext_type,
        help = "Numbers of the pages to render (defaults to all)",
    )
    parser.add_argument(
        "--scale",
        default = 1,
        type = float,
        help = "Define the resolution of the output images. By default, one PDF point (1/72in) is rendered to 1x1 pixel. This factor scales the number of pixels that represent one point.",
    )
    parser.add_argument(
        "--rotation",
        default = 0,
        type = int,
        choices = (0, 90, 180, 270),
        help = "Rotate pages by 90, 180 or 270 degrees",
    )
    parser.add_argument(
        "--fill-colour",
        default = (255, 255, 255, 255),
        help = "Colour the bitmap will be filled with before rendering. It shall be given in RGBA format as a sequence of integers ranging from 0 to 255. Defaults to white.",
        **ColourOpts,
    )
    parser.add_argument(
        "--force-halftone",
        action = "store_true",
        help = "Always use halftone for image stretching",
    )
    parser.add_argument(
        "--no-annotations",
        action = "store_true",
        help = "Prevent rendering of PDF annotations",
    )
    parser.add_argument(
        "--no-forms",
        action = "store_true",
        help = "Prevent rendering of PDF forms",
    )
    parser.add_argument(
        "--optimise-mode",
        default = pdfium.OptimiseMode.NONE,
        type = lambda string: pdfium.OptimiseMode[string.upper()],
        help = "Select a rendering optimisation mode (none, lcd_display, printing)",
    )
    parser.add_argument(
        "--greyscale",
        action = "store_true",
        help = "Whether to render in greyscale mode (no colours)",
    )
    parser.add_argument(
        "--crop",
        nargs = 4,
        type = float,
        default = (0, 0, 0, 0),
        help = "Amount to crop from (left, bottom, right, top)",
    )
    parser.add_argument(
        "--no-antialias",
        nargs = "+",
        default = (),
        choices = ("text", "image", "path"),
        help = "Item types that shall not be smoothed",
    )
    parser.add_argument(
        "--rev-byteorder",
        action = "store_true",
        help = "Render with reverse byte order internally, i. e. RGB(A) instead of BGR(A). The result should be completely identical.",
    )
    parser.add_argument(
        "--prefer-bgrx",
        action = "store_true",
        help = "Request the use of a four-channel pixel format for coloured output, even if rendering without transparency.",
    )
    parser.add_argument(
        "--processes",
        default = os.cpu_count(),
        type = int,
        help = "The number of processes to use for rendering (defaults to the number of CPU cores)",
    )
    
    colour_scheme = parser.add_argument_group(
        title = "Colour scheme",
        description = "Options for rendering with custom colour scheme",
    )
    colour_scheme.add_argument(
        "--path-fill",
        **ColourOpts
    )
    colour_scheme.add_argument(
        "--path-stroke",
        **ColourOpts
    )
    colour_scheme.add_argument(
        "--text-fill",
        **ColourOpts
    )
    colour_scheme.add_argument(
        "--text-stroke",
        **ColourOpts
    )
    colour_scheme.add_argument(
        "--fill-to-stroke",
        action = "store_true",
        help = "Whether fill paths need to be stroked.",
    )


def main(args):
    
    if not args.passwords:
        args.passwords = [None for _ in args.inputs]
    
    for input_path, password in zip(args.inputs, args.passwords):
        
        pdf = pdfium.PdfDocument(input_path, password=password)
        if args.pages:
            page_indices = args.pages
        else:
            page_indices = [i for i in range(len(pdf))]
        
        cs_kwargs = dict(
            path_fill = args.path_fill,
            path_stroke = args.path_stroke,
            text_fill = args.text_fill,
            text_stroke = args.text_stroke,
        )
        cs = None
        if all(cs_kwargs.values()):
            cs = pdfium.ColourScheme(
                fill_to_stroke = args.fill_to_stroke,
                **cs_kwargs,
            )
        elif any(cs_kwargs.values()):
            raise ValueError("If rendering with custom colour scheme, all parameters need to be set explicitly.")
        
        kwargs = dict(
            page_indices = page_indices,
            n_processes = args.processes,
            scale = args.scale,
            rotation = args.rotation,
            crop = args.crop,
            greyscale = args.greyscale,
            fill_colour = args.fill_colour,
            colour_scheme = cs,
            optimise_mode = args.optimise_mode,
            draw_annots = not args.no_annotations,
            draw_forms = not args.no_forms,
            force_halftone = args.force_halftone,
            rev_byteorder = args.rev_byteorder,
            prefer_bgrx = args.prefer_bgrx,
        )
        for type in args.no_antialias:
            kwargs["no_smooth%s" % type] = True
        
        prefix = splitext(basename(input_path))[0] + "_"
        n_digits = len(str( max(page_indices)+1 ))
        renderer = pdf.render_to(pdfium.BitmapConv.pil_image, **kwargs)
        
        for image, index in zip(renderer, page_indices):
            suffix = str(index+1).zfill(n_digits)
            output_path = "%s.%s" % (join(args.output, prefix+suffix), args.format)
            image.save(output_path)



# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

import enum
import ctypes
import pypdfium2._pypdfium as pdfium


class PdfiumError (RuntimeError):
    """ An exception from the PDFium library, detected by function return code. """
    pass


class FileAccess (enum.Enum):  # FIXME ambiguous name?
    """
    Different ways how files can be loaded.
    
    .. list-table:: Overview of file access modes
        :header-rows: 1
        :widths: auto
        
        * - Mode
          - PDFium loader
          - Comment
        * - :attr:`.NATIVE`
          - :func:`.FPDF_LoadDocument`
          - File access managed by PDFium in C/C++.
        * - :attr:`.BUFFER`
          - :func:`.FPDF_LoadCustomDocument`
          - Data read incrementally from Python file buffer.
        * - :attr:`.BYTES`
          - :func:`.FPDF_LoadMemDocument64`
          - Data loaded into memory and passed to PDFium at once.
    """
    NATIVE = 0
    BUFFER = 1
    BYTES  = 2


class OptimiseMode (enum.Enum):
    """ Modes defining how page rendering shall be optimised. """
    NONE = 0         #: No optimisation.
    LCD_DISPLAY = 1  #: Optimise for LCD displays (via subpixel rendering).
    PRINTING = 2     #: Optimise for printing.


class OutlineItem:
    """
    Class to store information about an entry in the table of contents ("bookmark").
    
    Parameters:
        level (int):
            Number of parent items.
        title (str):
            String of the bookmark.
        is_closed (bool):
            :data:`True` if child items shall be collapsed by default.
            :data:`False` if they shall be expanded by default.
            :data:`None` if the item has no descendants (i. e. *n_kids* == 0).
        n_kids (int):
            Number of child items (>= 0).
        page_index (int | None):
            Zero-based index of the page the bookmark points to.
            May be :data:`None` if the bookmark has no target page (or it could not be determined).
        view_mode (int):
            A view mode constant (:data:`PDFDEST_VIEW_*`) defining how the coordinates of *view_pos* shall be interpreted.
        view_pos (typing.Sequence[float]):
            Target position on the page the viewport should jump to when the bookmark is clicked.
            It is a sequence of :class:`float` values in PDF canvas units.
            Depending on *view_mode*, it can contain between 0 and 4 coordinates.
    """
    
    def __init__(
            self,
            level,
            title,
            is_closed,
            n_kids,
            page_index,
            view_mode,
            view_pos,
        ):
        self.level = level
        self.title = title
        self.is_closed = is_closed
        self.n_kids = n_kids
        self.page_index = page_index
        self.view_mode = view_mode
        self.view_pos = view_pos


def colour_tohex(colour, rev_byteorder):
    """
    Convert an RGBA colour specified by 4 integers ranging from 0 to 255 to a single 32-bit integer as required by PDFium.
    If using regular byte order, the output format will be ARGB. If using reversed byte order, it will be ABGR.
    """
    
    if len(colour) != 4:
        raise ValueError("Colour must consist of exactly 4 values.")
    if not all(0 <= c <= 255 for c in colour):
        raise ValueError("Colour value exceeds boundaries.")
    
    r, g, b, a = colour
    
    # colour is interpreted differently with FPDF_REVERSE_BYTE_ORDER (perhaps inadvertently?)
    if rev_byteorder:
        channels = (a, b, g, r)
    else:
        channels = (a, r, g, b)
    
    c_colour = 0
    shift = 24
    for c in channels:
        c_colour |= c << shift
        shift -= 8
    
    return c_colour


def get_functype(struct, funcname):
    """
    Parameters:
        struct (ctypes.Structure): A structure (e. g. ``FPDF_FILEWRITE``).
        funcname (str): Name of the callback function to implement (e. g. ``WriteBlock``).
    Returns:
        A :func:`ctypes.CFUNCTYPE` instance to wrap the callback function.
        For some reason, this is not done automatically, although the information is present in the bindings file.
        This is a convenience function to retrieve the declaration.
    """
    return {k: v for k, v in struct._fields_}[funcname]


class _reader_class:
    
    def __init__(self, buffer):
        self._buffer = buffer
    
    def __call__(self, _, position, p_buf, size):
        c_buf = ctypes.cast(p_buf, ctypes.POINTER(ctypes.c_char * size))
        self._buffer.seek(position)
        self._buffer.readinto(c_buf.contents)
        return 1


def is_input_buffer(maybe_buffer):
    """
    Returns:
        bool: True if the given object implements the methods ``seek()``, ``tell()``, ``read()``, and ``readinto()``. False otherwise.
    """
    return all( callable(getattr(maybe_buffer, a, None)) for a in ("seek", "tell", "read", "readinto") )


def get_fileaccess(buffer):
    """
    Acquire an :class:`FPDF_FILEACCESS` interface for a byte buffer.
    
    Returns:
        (FPDF_FILEACCESS, tuple): PDFium file access interface, and accompanying data that needs to be held in memory.
    """
    
    buffer.seek(0, 2)
    file_len = buffer.tell()
    buffer.seek(0)
    
    fileaccess = pdfium.FPDF_FILEACCESS()
    fileaccess.m_FileLen = file_len
    fileaccess.m_GetBlock = get_functype(pdfium.FPDF_FILEACCESS, "m_GetBlock")( _reader_class(buffer) )
    fileaccess.m_Param = None
    
    ld_data = (fileaccess.m_GetBlock, buffer)
    
    return fileaccess, ld_data


def _invert_dict(dictionary):
    """
    Returns:
        A copy of *dictionary*, with inverted keys and values.
    """
    return {v: k for k, v in dictionary.items()}


#: Get the number of channels for a PDFium pixel format constant.
BitmapTypeToNChannels = {
    pdfium.FPDFBitmap_Gray: 1,
    pdfium.FPDFBitmap_BGR:  3,
    pdfium.FPDFBitmap_BGRA: 4,
    pdfium.FPDFBitmap_BGRx: 4,
}

#: Convert a PDFium pixel format constant to string, assuming regular byte order.
BitmapTypeToStr = {
    pdfium.FPDFBitmap_Gray: "L",
    pdfium.FPDFBitmap_BGR:  "BGR",
    pdfium.FPDFBitmap_BGRA: "BGRA",
    pdfium.FPDFBitmap_BGRx: "BGRX",
}

#: Convert a reverse pixel format string to its regular counterpart.
BitmapStrReverseToRegular = {
    "BGR":  "RGB",
    "BGRA": "RGBA",
    "BGRX": "RGBX",
}

#: Convert a PDFium pixel format constant to string, assuming reversed byte order.
BitmapTypeToStrReverse = {
    pdfium.FPDFBitmap_Gray: "L",
    pdfium.FPDFBitmap_BGR:  "RGB",
    pdfium.FPDFBitmap_BGRA: "RGBA",
    pdfium.FPDFBitmap_BGRx: "RGBX",
}

#: Convert a PDFium view mode constant (:attr:`PDFDEST_VIEW_*`) to string.
ViewmodeToStr = {
    pdfium.PDFDEST_VIEW_XYZ:   "XYZ",
    pdfium.PDFDEST_VIEW_FIT:   "Fit",
    pdfium.PDFDEST_VIEW_FITH:  "FitH",
    pdfium.PDFDEST_VIEW_FITV:  "FitV",
    pdfium.PDFDEST_VIEW_FITR:  "FitR",
    pdfium.PDFDEST_VIEW_FITB:  "FitB",
    pdfium.PDFDEST_VIEW_FITBH: "FitBH",
    pdfium.PDFDEST_VIEW_FITBV: "FitBV",
    pdfium.PDFDEST_VIEW_UNKNOWN_MODE: "?",
}

#: Convert a PDFium error constant (:attr:`FPDF_ERR_*`) to string.
ErrorToStr = {
    pdfium.FPDF_ERR_SUCCESS:  "Success",
    pdfium.FPDF_ERR_UNKNOWN:  "Unknown error",
    pdfium.FPDF_ERR_FILE:     "File access error",
    pdfium.FPDF_ERR_FORMAT:   "Data format error",
    pdfium.FPDF_ERR_PASSWORD: "Incorrect password error",
    pdfium.FPDF_ERR_SECURITY: "Unsupported security scheme error",
    pdfium.FPDF_ERR_PAGE:     "Page not found or content error",
}

#: Convert a PDFium object type constant (:attr:`FPDF_PAGEOBJ_*`) to string.
ObjectTypeToStr = {
    pdfium.FPDF_PAGEOBJ_UNKNOWN: "unknown",
    pdfium.FPDF_PAGEOBJ_TEXT:    "text",
    pdfium.FPDF_PAGEOBJ_PATH:    "path",
    pdfium.FPDF_PAGEOBJ_IMAGE:   "image",
    pdfium.FPDF_PAGEOBJ_SHADING: "shading",
    pdfium.FPDF_PAGEOBJ_FORM:    "form",
}

#: Convert an object type string to a PDFium constant. Inversion of :data:`.ObjectTypeToStr`.
ObjectTypeToConst = _invert_dict(ObjectTypeToStr)

#: Convert a rotation value in degrees to a PDFium constant.
RotationToConst = {
    0:   0,
    90:  1,
    180: 2,
    270: 3,
}

#: Convert a PDFium rotation constant to a value in degrees. Inversion of :data:`.RotationToConst`.
RotationToDegrees = _invert_dict(RotationToConst)



# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

import math
import ctypes
import weakref
import logging
from ctypes import c_float
import pypdfium2._pypdfium as pdfium
from pypdfium2._helpers.misc import (
    OptimiseMode,
    PdfiumError,
    get_functype,
    colour_tohex,
    RotationToConst,
    RotationToDegrees,
    BitmapTypeToStr,
    BitmapTypeToStrReverse,
)
from pypdfium2._helpers.pageobject import (
    PdfPageObject,
)
from pypdfium2._helpers.converters import (
    BitmapConvBase,
    BitmapConvAliases,
)
from pypdfium2._helpers.textpage import PdfTextPage

try:
    import uharfbuzz as harfbuzz
except ImportError:
    harfbuzz = None

logger = logging.getLogger(__name__)


class PdfPage (BitmapConvAliases):
    """
    Page helper class.
    
    Attributes:
        raw (FPDF_PAGE): The underlying PDFium page handle.
        pdf (PdfDocument): Reference to the document this page belongs to.
    """
    
    def __init__(self, raw, pdf):
        self.raw = raw
        self.pdf = pdf
        # if the form env of the parent document is initialised, we could call FORM_OnAfterLoadPage() here
        self._finalizer = weakref.finalize(
            self, self._static_close,
            self.raw, self.pdf,
        )
    
    def _tree_closed(self):
        if self.raw is None:
            return True
        return self.pdf._tree_closed()
    
    @staticmethod
    def _static_close(raw, parent):
        # logger.debug("Closing page")
        if parent._tree_closed():
            logger.critical("Document closed before page (this is illegal). Document: %s" % parent)
        # if the form env of the parent document is initialised, we could call FORM_OnBeforeClosePage() here
        pdfium.FPDF_ClosePage(raw)
    
    def close(self):
        """
        Free memory by applying the finalizer for the underlying PDFium page.
        Please refer to the generic note on ``close()`` methods for details.
        """
        if self.raw is None:
            logger.warning("Duplicate close call suppressed on page %s" % self)
            return
        self._finalizer()
        self.raw = None
    
    
    def get_width(self):
        """
        Returns:
            float: Page width (horizontal size).
        """
        return pdfium.FPDF_GetPageWidthF(self.raw)
    
    def get_height(self):
        """
        Returns:
            float: Page height (vertical size).
        """
        return pdfium.FPDF_GetPageHeightF(self.raw)
    
    def get_size(self):
        """
        Returns:
            (float, float): Page width and height.
        """
        return (self.get_width(), self.get_height())
    
    def get_rotation(self):
        """
        Returns:
            int: Clockwise page rotation in degrees.
        """
        return RotationToDegrees[ pdfium.FPDFPage_GetRotation(self.raw) ]
    
    def set_rotation(self, rotation):
        """ Define the absolute, clockwise page rotation (0, 90, 180, or 270 degrees). """
        pdfium.FPDFPage_SetRotation(self.raw, RotationToConst[rotation])
    
    
    def _get_box(self, box_func, fallback_func):
        left, bottom, right, top = c_float(), c_float(), c_float(), c_float()
        success = box_func(self.raw, left, bottom, right, top)
        if not success:
            # TODO avoid repeated initialisation of c_float objects for fallback
            return fallback_func()
        return (left.value, bottom.value, right.value, top.value)
    
    def _set_box(self, box_func, l, b, r, t):
        if not all(isinstance(val, (int, float)) for val in (l, b, r, t)):
            raise ValueError("Box values must be int or float.")
        box_func(self.raw, l, b, r, t)
    
    def get_mediabox(self):
        """
        Returns:
            (float, float, float, float):
            The page MediaBox in PDF canvas units, consisting of four coordinates (usually x0, y0, x1, y1).
            Falls back to ANSI A (0, 0, 612, 792) in case MediaBox is not defined.
        """
        return self._get_box(pdfium.FPDFPage_GetMediaBox, lambda: (0, 0, 612, 792))
    
    def set_mediabox(self, l, b, r, t):
        """
        Set the page's MediaBox by passing four :class:`float` coordinates (usually x0, y0, x1, y1).
        """
        self._set_box(pdfium.FPDFPage_SetMediaBox, l, b, r, t)
    
    def get_cropbox(self):
        """
        Returns:
            The page's CropBox (If not defined, falls back to MediaBox).
        """
        return self._get_box(pdfium.FPDFPage_GetCropBox, self.get_mediabox)
    
    def set_cropbox(self, l, b, r, t):
        """ Set the page's CropBox. """
        self._set_box(pdfium.FPDFPage_SetCropBox, l, b, r, t)
    
    def get_bleedbox(self):
        """
        Returns:
            The page's BleedBox (If not defined, falls back to CropBox).
        """
        return self._get_box(pdfium.FPDFPage_GetBleedBox, self.get_cropbox)
    
    def set_bleedbox(self, l, b, r, t):
        """ Set the page's BleedBox. """
        self._set_box(pdfium.FPDFPage_SetBleedBox, l, b, r, t)
    
    def get_trimbox(self):
        """
        Returns:
            The page's TrimBox (If not defined, falls back to CropBox).
        """
        return self._get_box(pdfium.FPDFPage_GetTrimBox, self.get_cropbox)
    
    def set_trimbox(self, l, b, r, t):
        """ Set the page's TrimBox. """
        self._set_box(pdfium.FPDFPage_SetTrimBox, l, b, r, t)
    
    def get_artbox(self):
        """
        Returns:
            The page's ArtBox (If not defined, falls back to CropBox).
        """
        return self._get_box(pdfium.FPDFPage_GetArtBox, self.get_cropbox)
    
    def set_artbox(self, l, b, r, t):
        """ Set the page's ArtBox. """
        self._set_box(pdfium.FPDFPage_SetArtBox, l, b, r, t)
    
    
    def get_textpage(self):
        """
        Returns:
            PdfTextPage: The text page that corresponds to this page.
        """
        textpage = pdfium.FPDFText_LoadPage(self.raw)
        if not textpage:
            raise PdfiumError("Loading the text page failed")
        return PdfTextPage(textpage, self)
    
    
    def insert_object(self, pageobj):
        """
        Insert a page object into the page.
        
        Position and form are defined by the object's matrix.
        If it is the identity matrix, the object will appear as-is on the bottom left corner of the page.
        
        The page object must not belong to a page yet. If it belongs to a PDF, this page must be part of the PDF.
        
        You may want to call :meth:`.generate_content` once you finished adding new content to the page.
        
        Parameters:
            pageobj (PdfPageObject): The page object to insert.
        """
        
        if pageobj.page is not None:
            raise ValueError("The pageobject you attempted to insert already belongs to a page.")
        if (pageobj.pdf is not None) and (pageobj.pdf is not self.pdf):
            raise ValueError("The pageobject you attempted to insert belongs to a different PDF.")
        
        pdfium.FPDFPage_InsertObject(self.raw, pageobj.raw)
        
        pageobj.page = self
        pageobj.pdf = self.pdf
    
    
    def generate_content(self):
        """
        Generate added page content.
        This function shall be called to apply changes before saving the document or reloading the page.
        """
        success = pdfium.FPDFPage_GenerateContent(self.raw)
        if not success:
            raise PdfiumError("Generating page content failed.")
    
    
    def insert_text(
            self,
            text,
            pos_x,
            pos_y,
            font_size,
            hb_font,
            pdf_font,
        ):
        """
        *Requires* :mod:`uharfbuzz`
        
        Insert text into the page at a specified position, using the writing system's ligature.
        This function supports Asian scripts such as Hindi.
        
        You may want to call :meth:`.generate_content` once you finished adding new content to the page.
        
        Parameters:
            text (str):
                The message to insert.
            pos_x (float):
                Distance from left border to first character.
            pos_y (float):
                Distance from bottom border to first character.
            font_size (float):
                Font size the text shall have.
            hb_font (HarfbuzzFont):
                Harfbuzz font data.
            pdf_font (PdfFont):
                PDF font data.
        """
        
        # User-contributed code
        # SPDX-FileCopyrightText: 2022 Anurag Bansal <anurag.bansal.585@gmail.com>
        
        hb_buffer = harfbuzz.Buffer()
        hb_buffer.add_str(text)
        hb_buffer.guess_segment_properties()
        hb_features = {"kern": True, "liga": True}
        harfbuzz.shape(hb_font.font, hb_buffer, hb_features)
        
        start_point = pos_x
        for info, pos in zip(hb_buffer.glyph_infos, hb_buffer.glyph_positions):
            pdf_textobj = pdfium.FPDFPageObj_CreateTextObj(self.pdf.raw, pdf_font.raw, font_size)
            pdfium.FPDFText_SetCharcodes(pdf_textobj, ctypes.c_uint32(info.codepoint), 1)
            pdfium.FPDFPageObj_Transform(
                pdf_textobj,
                1, 0, 0, 1,
                start_point - (pos.x_offset / hb_font.scale) * font_size,
                pos_y,
            )
            pdfium.FPDFPage_InsertObject(self.raw, pdf_textobj)
            start_point += (pos.x_advance / hb_font.scale) * font_size
    
    
    def get_objects(self, max_depth=2, form=None, level=0):
        """
        Iterate through the page objects on this page.
        
        Parameters:
            max_depth (int):
                Maximum recursion depth to consider when descending into Form XObjects.
        
        Yields:
            :class:`.PdfPageObject`: The page object.
        """
        
        if form is None:
            count_objects = pdfium.FPDFPage_CountObjects
            get_object = pdfium.FPDFPage_GetObject
            parent = self.raw
        else:
            count_objects = pdfium.FPDFFormObj_CountObjects
            get_object = pdfium.FPDFFormObj_GetObject
            parent = form
        
        n_objects = count_objects(parent)
        if n_objects == 0:
            return
        elif n_objects < 0:
            raise PdfiumError("Failed to get number of page objects.")
        
        for i in range(n_objects):
            
            raw_obj = get_object(parent, i)
            if raw_obj is None:
                raise PdfiumError("Failed to get page object.")
            
            type = pdfium.FPDFPageObj_GetType(raw_obj)
            yield PdfPageObject(
                raw = raw_obj,
                type = type,
                page = self,
                pdf = self.pdf,
                level = level,
            )
            
            if level < max_depth-1 and type == pdfium.FPDF_PAGEOBJ_FORM:
                yield from self.get_objects(
                    max_depth = max_depth,
                    form = raw_obj,
                    level = level + 1,
                )
    
    
    def render_to(self, converter, **renderer_kws):
        """
        Rasterise a page to a specific output format.
        
        Parameters:
            converter (BitmapConvBase | typing.Callable):
                A translator to convert the output of :meth:`.render_base`. See :class:`.BitmapConv` for a set of built-in converters.
            renderer_kws (dict):
                Keyword arguments to the renderer.
        
        Returns:
            typing.Any: Converter-specific result.
        
        Examples:
            
            .. code-block:: python
                
                # convert to a NumPy array
                array, cl_format = render_to(BitmapConv.numpy_ndarray, ...)
                # passing an initialised converter without arguments is equivalent
                array, cl_format = render_to(BitmapConv.numpy_ndarray(), ...)
                
                # convert to a PIL image (with default settings)
                image = render_to(BitmapConv.pil_image, ...)
                
                # convert to PIL image (with specific settings)
                image = render_to(BitmapConv.pil_image(prefer_la=True), ...)
                
                # convert to bytes using the special "any" converter factory
                data, cl_format, size = render_to(BitmapConv.any(bytes), ...)
        """
        
        args = (self.render_base(**renderer_kws), renderer_kws)
        if isinstance(converter, BitmapConvBase):
            return converter.run(*args, *converter.args, **converter.kwargs)
        elif isinstance(converter, type) and issubclass(converter, BitmapConvBase):
            return converter().run(*args)
        elif callable(converter):
            return converter(*args)
        else:
            raise ValueError("Converter must be an instance or subclass of BitmapConvBase, or a callable, but %s was given." % converter)
    
    
    def render_base(
            self,
            scale = 1,
            rotation = 0,
            crop = (0, 0, 0, 0),
            greyscale = False,
            fill_colour = (255, 255, 255, 255),
            colour_scheme = None,
            optimise_mode = OptimiseMode.NONE,
            draw_annots = True,
            draw_forms = True,
            no_smoothtext = False,
            no_smoothimage = False,
            no_smoothpath = False,
            force_halftone = False,
            limit_image_cache = False,
            rev_byteorder = False,
            prefer_bgrx = False,
            force_bitmap_format = None,
            extra_flags = 0,
            allocator = None,
            memory_limit = 2**30,
        ):
        """
        Rasterise the page to a :class:`ctypes.c_ubyte` array. This is the base method for :meth:`.render_to`.
        
        Parameters:
            
            scale (float):
                A factor scaling the number of pixels that represent the length of 1 PDF canvas unit (usually 1/72 in). [1]_
                This defines the resolution of the image. To convert a DPI value to a scale factor, multiply it by the size of 1 canvas unit in inches.
                
                .. [1] Since PDF 1.6, pages may define a so-called user unit. In this case, 1 canvas unit is equivalent to ``user_unit * (1/72)`` inches. pypdfium2 currently does not take this into account.
                
            rotation (int):
                A rotation value in degrees (0, 90, 180, or 270), in addition to page rotation.
            
            crop (typing.Tuple[float, float, float, float]):
                Amount in PDF canvas units to cut off from page borders (left, bottom, right, top).
                Crop is applied after rotation.
            
            greyscale (bool):
                Whether to render in greyscale mode (no colours).
            
            fill_colour (typing.Tuple[int, int, int, int]):
                Colour the bitmap will be filled with before rendering.
                Shall be a list of values for red, green, blue and alpha, ranging from 0 to 255.
                For RGB, 0 will include nothing of the colour in question, while 255 will fully include it.
                For Alpha, 0 means full transparency, while 255 means no transparency.
            
            colour_scheme (ColourScheme | None):
                A custom colour scheme for rendering, defining fill and stroke colours for path and text objects.
            
            optimise_mode (OptimiseMode):
                How to optimise page rendering.
            
            draw_annots (bool):
                Whether to render page annotations.
            
            draw_forms (bool):
                Whether to render form fields.
            
            no_smoothtext (bool):
                Disable anti-aliasing of text. Implicitly wipes out :attr:`.OptimiseMode.LCD_DISPLAY`.
            
            no_smoothimage (bool):
                Disable anti-aliasing of images.
            
            no_smoothpath (bool):
                Disable anti-aliasing of paths.
            
            force_halftone (bool):
                Always use halftone for image stretching.
            
            limit_image_cache (bool):
                Limit image cache size.
            
            rev_byteorder (bool):
                By default, the output pixel format will be ``BGR(A/X)``.
                This option may be used to render with reversed byte order, leading to ``RGB(A/X)`` output instead.
                ``L`` is unaffected.
            
            prefer_bgrx (bool):
                Request the use of a four-channel pixel format for coloured output, even if rendering without transparency.
                (i. e. ``BGRX``/``RGBX`` rather than ``BGR``/``RGB``).
            
            force_bitmap_format (int | None):
                If given, override the automatic pixel format selection and enforce the use of a specific format.
                May be one of the :attr:`FPDFBitmap_*` constants, except :attr:`FPDFBitmap_Unknown`.
                For instance, this may be used to render in greyscale mode while using ``BGR`` as output format (default choice would be ``L``).
            
            extra_flags (int):
                Additional PDFium rendering flags. Multiple flags may be combined with bitwise OR (``|`` operator).
            
            allocator (typing.Callable | None):
                A function to provide a custom ctypes buffer. It is called with the required buffer size in bytes.
                If not given, a new :class:`ctypes.c_ubyte` array is allocated by Python (this simplify memory management, as opposed to allocation by PDFium).
            
            memory_limit (int | None):
                Maximum number of bytes that may be allocated (defaults to 1 GiB rsp. 2^30 bytes).
                If the limit would be exceeded, a :exc:`RuntimeError` is raised.
                If :data:`None` or 0, this function may allocate arbitrary amounts of memory as far as Python and the OS permit.
            
        Returns:
            (ctypes array, str, (int, int)): Bitmap data, colour format, and image size.
            The colour format may be ``BGR``/``RGB``, ``BGRA``/``RGBA``, ``BGRX``/``RGBX``, or ``L``, depending on the parameters *colour*, *greyscale*, *rev_byteorder* and *prefer_bgrx*.
            Image size is given in pixels as a tuple of width and height.
        """
        
        if force_bitmap_format in (None, pdfium.FPDFBitmap_Unknown):
            cl_pdfium = _auto_bitmap_format(fill_colour, greyscale, prefer_bgrx)
        else:
            cl_pdfium = force_bitmap_format
        
        if cl_pdfium == pdfium.FPDFBitmap_Gray:
            rev_byteorder = False
        
        if rev_byteorder:
            cl_string = BitmapTypeToStrReverse[cl_pdfium]
        else:
            cl_string = BitmapTypeToStr[cl_pdfium]
        
        c_fill_colour = colour_tohex(fill_colour, rev_byteorder)
        n_channels = len(cl_string)
        
        src_width  = math.ceil(self.get_width()  * scale)
        src_height = math.ceil(self.get_height() * scale)
        if rotation in (90, 270):
            src_width, src_height = src_height, src_width
        
        crop = [math.ceil(c*scale) for c in crop]
        width  = src_width  - crop[0] - crop[2]
        height = src_height - crop[1] - crop[3]
        if any(d < 1 for d in (width, height)):
            raise ValueError("Crop exceeds page dimensions (in px): width %s, height %s, crop %s" % (src_width, src_height, crop))
        
        stride = width * n_channels
        n_bytes = stride * height
        if memory_limit and n_bytes > memory_limit:
            raise RuntimeError(
                "Planned allocation of %s bytes exceeds the defined limit of %s. " % (n_bytes, memory_limit) +
                "Consider adjusting the *memory_limit* parameter."
            )
        
        if allocator is None:
            buffer = (ctypes.c_ubyte * n_bytes)()
        else:
            buffer = allocator(n_bytes)
            if ctypes.sizeof(buffer) < n_bytes:
                raise RuntimeError("Not enough bytes allocated (buffer length: %s, required bytes: %s)." % (ctypes.sizeof(buffer), n_bytes))
        
        bitmap = pdfium.FPDFBitmap_CreateEx(width, height, cl_pdfium, buffer, stride)
        pdfium.FPDFBitmap_FillRect(bitmap, 0, 0, width, height, c_fill_colour)
        
        render_flags = extra_flags
        if greyscale:
            render_flags |= pdfium.FPDF_GRAYSCALE
        if draw_annots:
            render_flags |= pdfium.FPDF_ANNOT
        if no_smoothtext:
            render_flags |= pdfium.FPDF_RENDER_NO_SMOOTHTEXT
        if no_smoothimage:
            render_flags |= pdfium.FPDF_RENDER_NO_SMOOTHIMAGE
        if no_smoothpath:
            render_flags |= pdfium.FPDF_RENDER_NO_SMOOTHPATH
        if force_halftone:
            render_flags |= pdfium.FPDF_RENDER_FORCEHALFTONE
        if limit_image_cache:
            render_flags |= pdfium.FPDF_RENDER_LIMITEDIMAGECACHE
        if rev_byteorder:
            render_flags |= pdfium.FPDF_REVERSE_BYTE_ORDER
        if colour_scheme and colour_scheme.fill_to_stroke:
            render_flags |= pdfium.FPDF_CONVERT_FILL_TO_STROKE
        
        if optimise_mode is OptimiseMode.NONE:
            pass
        elif optimise_mode is OptimiseMode.LCD_DISPLAY:
            render_flags |= pdfium.FPDF_LCD_TEXT
        elif optimise_mode is OptimiseMode.PRINTING:
            render_flags |= pdfium.FPDF_PRINTING
        else:
            raise ValueError("Invalid optimise_mode %s" % optimise_mode)
        
        render_args = (bitmap, self.raw, -crop[0], -crop[3], src_width, src_height, RotationToConst[rotation], render_flags)
        
        if colour_scheme is None:
            pdfium.FPDF_RenderPageBitmap(*render_args)
        else:
            
            ifsdk_pause = pdfium.IFSDK_PAUSE()
            ifsdk_pause.version = 1
            ifsdk_pause.NeedToPauseNow = get_functype(pdfium.IFSDK_PAUSE, "NeedToPauseNow")(lambda _: False)
            
            fpdf_cs = colour_scheme.convert(rev_byteorder)
            status = pdfium.FPDF_RenderPageBitmapWithColorScheme_Start(*render_args, fpdf_cs, ifsdk_pause)
            
            assert status == pdfium.FPDF_RENDER_DONE
            pdfium.FPDF_RenderPage_Close(self.raw)
        
        if draw_forms:
            form_type = pdfium.FPDF_GetFormType(self.pdf.raw)  # consider moving to document ?
            if form_type != pdfium.FORMTYPE_NONE:
                form_env = self.pdf.init_formenv()
                pdfium.FPDF_FFLDraw(form_env, *render_args)
        
        return buffer, cl_string, (width, height)


def _auto_bitmap_format(fill_colour, greyscale, prefer_bgrx):
    # no need to take alpha values of colour_scheme into account (drawings are additive)
    if (fill_colour[3] < 255):
        return pdfium.FPDFBitmap_BGRA
    elif greyscale:
        return pdfium.FPDFBitmap_Gray
    elif prefer_bgrx:
        return pdfium.FPDFBitmap_BGRx
    else:
        return pdfium.FPDFBitmap_BGR


class ColourScheme:
    """
    Rendering colour scheme.
    Each colour shall be provided as a list of values for red, green, blue and alpha, ranging from 0 to 255.
    """
    
    def __init__(
            self,
            path_fill,
            path_stroke,
            text_fill,
            text_stroke,
            fill_to_stroke = False,
        ):
        self.colours = dict(
            path_fill_color = path_fill,
            path_stroke_color = path_stroke,
            text_fill_color = text_fill,
            text_stroke_color = text_stroke,
        )
        self.fill_to_stroke = fill_to_stroke
    
    def convert(self, rev_byteorder):
        """
        Returns:
            The colour scheme as :class:`FPDF_COLORSCHEME` object.
        """
        fpdf_cs = pdfium.FPDF_COLORSCHEME()
        for key, value in self.colours.items():
            setattr(fpdf_cs, key, colour_tohex(value, rev_byteorder))
        return fpdf_cs



# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

import ctypes
import weakref
import logging
from ctypes import c_double
import pypdfium2._pypdfium as pdfium
from pypdfium2._helpers.misc import PdfiumError

logger = logging.getLogger(__name__)


class PdfTextPage:
    """
    Text page helper class.
    
    Attributes:
        raw (FPDF_TEXTPAGE): The underlying PDFium textpage handle.
        page (PdfPage): Reference to the page this textpage belongs to.
        n_chars (int): Number of characters on the page, at the time of initialisation.
    """
    
    def __init__(self, raw, page):
        self.raw = raw
        self.page = page
        self._finalizer = weakref.finalize(
            self, self._static_close,
            self.raw, self.page,
        )
        self.n_chars = pdfium.FPDFText_CountChars(self.raw)
    
    def _tree_closed(self):
        if self.raw is None:
            return True
        return self.page._tree_closed()
    
    @staticmethod
    def _static_close(raw, parent):
        # logger.debug("Closing text page")
        if parent._tree_closed():
            logger.critical("Some parent closed before text page (this is illegal). Direct parent: %s" % parent)
        pdfium.FPDFText_ClosePage(raw)
    
    def close(self):
        """
        Free memory by applying the finalizer for the underlying PDFium text page.
        Please refer to the generic note on ``close()`` methods for details.
        """
        if self.raw is None:
            logger.warning("Duplicate close call suppressed on text page %s" % self)
            return
        self._finalizer()
        self.raw = None
    
    
    def count_chars(self):
        """
        Deprecated alias for :attr:`.n_chars`. Will be removed with the next major release.
        """
        return self.n_chars
    
    
    @staticmethod
    def _check_span(n_chars, index, count):
        if not (0 <= index < index+count <= n_chars):
            raise ValueError("Character span is out of bounds.")
    
    
    def get_text_range(self, index=0, count=0, errors="ignore"):
        """
        Extract text from a given range.
        
        See `this benchmark <https://github.com/py-pdf/benchmarks>`_ for a performance and quality comparison with other tools.
        
        Parameters:
            index (int): Index of the first character to include.
            count (int): Number of characters to be extracted. If 0, it defaults to the number of characters on the page minus *index*.
            errors (str): Error treatment when decoding the data (see :meth:`bytes.decode`).
        Returns:
            str: The text in the range in question, or an empty string if no text was found.
        """
        
        if self.n_chars == 0:
            return ""
        if count == 0:
            count = self.n_chars - index
        self._check_span(self.n_chars, index, count)
        
        n_bytes = count * 2
        buffer = ctypes.create_string_buffer(n_bytes+2)
        buffer_ptr = ctypes.cast(buffer, ctypes.POINTER(ctypes.c_ushort))
        pdfium.FPDFText_GetText(self.raw, index, count, buffer_ptr)
        return buffer.raw[:n_bytes].decode("utf-16-le", errors=errors)
    
    
    def get_text_bounded(self, left=None, bottom=None, right=None, top=None, errors="ignore"):
        """
        Extract text from given boundaries in PDF coordinates.
        If a parameter is :data:`None`, it defaults to the corresponding CropBox value.
        
        Parameters:
            errors (str): Error treatment when decoding the data (see :meth:`bytes.decode`).
        Returns:
            str: The text on the page area in question, or an empty string if no text was found.
        """
        
        if self.n_chars == 0:
            return ""
        
        cropbox = self.page.get_cropbox()
        if left is None:
            left = cropbox[0]
        if bottom is None:
            bottom = cropbox[1]
        if right is None:
            right = cropbox[2]
        if top is None:
            top = cropbox[3]
        
        args = (self.raw, left, top, right, bottom)
        n_chars = pdfium.FPDFText_GetBoundedText(*args, None, 0)
        if n_chars <= 0:
            return ""
        
        n_bytes = 2 * n_chars
        buffer = ctypes.create_string_buffer(n_bytes)
        buffer_ptr = ctypes.cast(buffer, ctypes.POINTER(ctypes.c_ushort))
        pdfium.FPDFText_GetBoundedText(*args, buffer_ptr, n_chars)
        return buffer.raw.decode("utf-16-le", errors=errors)
    
    
    def get_text(self, *args, **kwargs):
        """
        Deprecated alias for :meth:`.get_text_bounded`. Will be removed with the next major release.
        """
        return self.get_text_bounded(*args, **kwargs)
    
    
    def count_rects(self, index=0, count=0):
        """
        Parameters:
            index (int): Character index at which to start.
            count (int): Character count to consider (defaults to :meth:`.count_chars`).
        Returns:
            int: The number of text rectangles on the page.
        """
        
        if self.n_chars == 0:
            return 0
        if count == 0:
            count = self.n_chars
        self._check_span(self.n_chars, index, count)
        
        return pdfium.FPDFText_CountRects(self.raw, index, count)
    
    
    def get_index(self, x, y, x_tol, y_tol):
        """
        Get the character index for a given position.
        
        Parameters:
            x (float): Horizontal position (in PDF canvas units).
            y (float): Vertical position.
            x_tol (float): Horizontal tolerance.
            y_tol (float): Vertical tolerance.
        Returns:
            int | None: The index of the character at or nearby the point (x, y).
            May be :data:`None` if there is no character or an error occurred.
        """
        index = pdfium.FPDFText_GetCharIndexAtPos(self.raw, x, y, x_tol, y_tol)
        if index < 0:
            return None
        return index
    
    
    def get_charbox(self, index, loose=False):
        """
        Get the bounding box of a single character.
        
        Parameters:
            index (int):
                Slot of the character to work with, in the page's character array.
            loose (bool):
                If True, the entire glyph bounds will be covered, without taking the actual glyph shape into account.
        Returns:
            (float, float, float, float):
            Values for left, bottom, right and top in PDF canvas units.
        """
        
        if not 0 <= index < self.n_chars:
            raise ValueError("Character index %s is out of bounds. The maximum index is %d." % (index, self.n_chars-1))
        
        if loose:
            rect = pdfium.FS_RECTF()
            success = pdfium.FPDFText_GetLooseCharBox(self.raw, index, rect)
            left, bottom, right, top = rect.left, rect.bottom, rect.right, rect.top
        else:
            left, bottom, right, top = c_double(), c_double(), c_double(), c_double()
            success = pdfium.FPDFText_GetCharBox(self.raw, index, left, right, bottom, top)
            left, bottom, right, top = left.value, bottom.value, right.value, top.value
        
        if not success:
            raise PdfiumError("Retrieving the char box failed")
        
        return left, bottom, right, top
    
    
    def get_rectboxes(self, index=0, count=0):
        """
        Get the bounding boxes of text rectangles in the requested scope.
        
        Yields:
            Coordinates for left, bottom, right, and top (as :class:`float` values).
        """
        n_rects = self.count_rects(index, count)
        for index in range(n_rects):
            left, top, right, bottom = c_double(), c_double(), c_double(), c_double()
            pdfium.FPDFText_GetRect(self.raw, index, left, top, right, bottom)
            yield (left.value, bottom.value, right.value, top.value)
    
    
    def get_links(self):
        """
        Iterate through web links on the page.
        
        Yields:
            :class:`str`: A web link string.
        """
        
        links = pdfium.FPDFLink_LoadWebLinks(self.raw)
        n_links = pdfium.FPDFLink_CountWebLinks(links)
        
        for i in range(n_links):
            n_chars = pdfium.FPDFLink_GetURL(links, i, None, 0)
            n_bytes = n_chars * 2
            buffer = ctypes.create_string_buffer(n_bytes)
            buffer_ptr = ctypes.cast(buffer, ctypes.POINTER(ctypes.c_ushort))
            pdfium.FPDFLink_GetURL(links, i, buffer_ptr, n_chars)
            yield buffer.raw[:n_bytes-2].decode("utf-16-le")
        
        pdfium.FPDFLink_CloseWebLinks(links)
    
    
    def search(self, text, index=0, match_case=False, match_whole_word=False):
        """
        Locate text on the page.
        
        Parameters:
            text (str):
                The string to search for.
            index (int):
                Character index at which to start searching.
            match_case (bool):
                If :data:`True`, the search will be case-specific (upper and lower letters treated as different characters).
            match_whole_word (bool):
                If :data:`True`, substring occurrences will be ignored (e. g. `cat` would not match `category`).
        Returns:
            PdfTextSearcher: A helper object to search text.
        """
        
        if len(text) == 0:
            raise ValueError("Text length must be >0.")
        
        flags = 0
        if match_case:
            flags |= pdfium.FPDF_MATCHCASE
        if match_whole_word:
            flags |= pdfium.FPDF_MATCHWHOLEWORD
        
        # assuming the pointer returned by ctypes.cast() keeps the casted object alive
        enc_text = (text + "\x00").encode("utf-16-le")
        enc_text_ptr = ctypes.cast(enc_text, ctypes.POINTER(ctypes.c_ushort))
        search = pdfium.FPDFText_FindStart(self.raw, enc_text_ptr, flags, index)
        return PdfTextSearcher(search, self)


class PdfTextSearcher:
    """
    Text searcher helper class.
    
    Attributes:
        raw (FPDF_SCHHANDLE): The underlying PDFium searcher handle.
        textpage (PdfTextPage): Reference to the textpage this searcher belongs to.
    """
    
    def __init__(self, raw, textpage):
        self.raw = raw
        self.textpage = textpage
        self._finalizer = weakref.finalize(
            self, self._static_close,
            self.raw, self.textpage,
        )
    
    def _tree_closed(self):
        if self.raw is None:
            return True
        return self.textpage._tree_closed()
    
    @staticmethod
    def _static_close(raw, parent):
        # logger.debug("Closing text searcher")
        if parent._tree_closed():
            logger.critical("Some parent closed before text searcher (this is illegal). Direct parent: %s" % parent)
        pdfium.FPDFText_FindClose(raw)
    
    def close(self):
        """
        Free memory by applying the finalizer for the underlying PDFium text searcher.
        Please refer to the generic note on ``close()`` methods for details.
        """
        if self.raw is None:
            logger.warning("Duplicate close call suppressed on text searcher %s" % self)
            return
        self._finalizer()
        self.raw = None
    
    def _get_occurrence(self, find_func):
        found = find_func(self.raw)
        if not found:
            return None
        index = pdfium.FPDFText_GetSchResultIndex(self.raw)
        count = pdfium.FPDFText_GetSchCount(self.raw)
        return tuple( [box for box in self.textpage.get_rectboxes(index, count)] )
    
    def get_next(self):
        """
        Returns:
            typing.Sequence[ typing.Tuple[float, float, float, float] ]:
            A list of bounding boxes for the next occurrence, or :data:`None` if the last occurrence was passed.
        """
        return self._get_occurrence(pdfium.FPDFText_FindNext)
    
    def get_prev(self):
        """
        Returns:
            typing.Sequence[ typing.Tuple[float, float, float, float] ]:
            A list of bounding boxes for the previous occurrence (i. e. the one before the last valid occurrence),
            or :data:`None` if the first occurrence was passed.
        """
        return self._get_occurrence(pdfium.FPDFText_FindPrev)



# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

import math
import pypdfium2._pypdfium as pdfium


class PdfMatrix:
    """
    PDF transformation matrix helper class (Python).
    
    See the PDF 1.7 specification, Section 8.3.3 ("Common Transformations").
    
    Note:
        * The PDF format uses row vectors.
        * Transformations operate from the origin of the coordinate system.
    
    Attributes:
        a (float): Matrix value [0][0].
        b (float): Matrix value [0][1].
        c (float): Matrix value [1][0].
        d (float): Matrix value [1][1].
        e (float): Matrix value [2][0] (X translation).
        f (float): Matrix value [2][1] (Y translation).
    """
    
    # The effect of applying the matrix on a vector (x, y) is (ax+cy+e, bx+dy+f)
    
    def __init__(self, a=1, b=0, c=0, d=1, e=0, f=0):
        self.set(a, b, c, d, e, f)
    
    def __eq__(self, matrix):
        if type(self) is not type(matrix):
            return False
        return (self.get() == matrix.get())
    
    def __repr__(self):
        return "PdfMatrix%s" % (self.get(), )
    
    def get(self):
        """
        Get the matrix as tuple of the form (a, b, c, d, e, f).
        """
        return (self.a, self.b, self.c, self.d, self.e, self.f)
    
    def set(self, a, b, c, d, e, f):
        """
        Set the matrix values.
        """
        self.a = a
        self.b = b
        self.c = c
        self.d = d
        self.e = e
        self.f = f
    
    def copy(self):
        """
        Returns:
            An independent copy of the matrix.
        """
        return PdfMatrix(*self.get())
    
    @classmethod
    def from_pdfium(cls, fs_matrix):
        """
        Load a :class:`.PdfMatrix` from a raw :class:`FS_MATRIX` object.
        """
        return PdfMatrix(
            fs_matrix.a,
            fs_matrix.b,
            fs_matrix.c,
            fs_matrix.d,
            fs_matrix.e,
            fs_matrix.f,
        )
    
    def to_pdfium(self):
        """
        Convert the matrix to a raw :class:`FS_MATRIX` object.
        """
        return pdfium.FS_MATRIX(*self.get())
    
    def multiply(self, other):
        """
        Multiply this matrix by another :class:`.PdfMatrix`, to concatenate transformations.
        """
        # M1 x M2 (self x other)
        # (a1, b1, 0)   (a2, b2, 0)   (a1a2+b1c2,    a1b2+b1d2,    0)
        # (c1, d1, 0) x (c2, d2, 0) = (c1a2+d1c2,    c1b2+d1d2,    0)
        # (e1, f1, 1)   (e2, f2, 0)   (e1a2+f1c2+e2, e1b2+f1d2+f2, 1)
        new_matrix = (
            self.a*other.a + self.b*other.c,            # a
            self.a*other.b + self.b*other.d,            # b
            self.c*other.a + self.d*other.c,            # c
            self.c*other.b + self.d*other.d,            # d
            self.e*other.a + self.f*other.c + other.e,  # e
            self.e*other.b + self.f*other.d + other.f,  # f
        )
        self.set(*new_matrix)
    
    def translate(self, x, y):
        """
        Parameters:
            x (float): Horizontal shift (<0: left, >0: right).
            y (float): Vertical shift.
        """
        # same as self.multiply( PdfMatrix(1, 0, 0, 1, x, y) )
        self.e += x
        self.f += y
    
    def scale(self, x, y):
        """
        Parameters:
            x (float): A factor to scale the X axis (<1: compress, >1: stretch).
            y (float): A factor to scale the Y axis.
        """
        # same as a*=x, b*=y, c*=x, d*=y, e*=x, f*=y
        self.multiply( PdfMatrix(x, 0, 0, y) )
    
    def rotate(self, angle):
        """
        Parameters:
            angle (float): Clockwise angle in degrees to rotate the matrix.
        """
        # row vectors -> b = -s leads to clockwise rotation indeed
        angle = (angle/180) * math.pi  # arc measure
        c, s = math.cos(angle), math.sin(angle)
        self.multiply( PdfMatrix(c, -s, s, c) )
    
    def mirror(self, vertical, horizontal):
        """
        Parameters:
            vertical (bool): Whether to mirror at the Y axis.
            horizontal (bool): Whether to mirror at the X axis.
        """
        s_x = (-1 if vertical else 1)
        s_y = (-1 if horizontal else 1)
        self.scale(s_x, s_y)
    
    def skew(self, x_angle, y_angle):
        """
        Parameters:
            x_angle (float): Inner angle in degrees to skew the X axis.
            y_angle (float): Inner angle in degrees to skew the Y axis.
        """
        tan_a = math.tan((x_angle/180) * math.pi)
        tan_b = math.tan((y_angle/180) * math.pi)
        self.multiply( PdfMatrix(1, tan_a, tan_b, 1) )



# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

import io
import os
import os.path
import weakref
import ctypes
import logging
import functools
from concurrent.futures import ProcessPoolExecutor

import pypdfium2._pypdfium as pdfium
from pypdfium2._helpers.misc import (
    OutlineItem,
    FileAccess,
    PdfiumError,
    ErrorToStr,
    ViewmodeToStr,
    get_functype,
    get_fileaccess,
    is_input_buffer,
)
from pypdfium2._helpers.pageobject import (
    PdfPageObject,
)
from pypdfium2._helpers.converters import BitmapConvAliases
from pypdfium2._helpers.page import PdfPage

try:
    import uharfbuzz as harfbuzz
except ImportError:
    harfbuzz = None

logger = logging.getLogger(__name__)


class PdfDocument (BitmapConvAliases):
    """
    Document helper class.
    
    Parameters:
        input_data (str | bytes | typing.BinaryIO | FPDF_DOCUMENT):
            The input PDF given as file path, bytes, byte buffer, or raw PDFium document handle.
            :func:`.is_input_buffer` defines which objects are recognised as byte buffers.
        password (str | bytes):
            A password to unlock the PDF, if encrypted.
            If the document is not encrypted but a password was given, PDFium will ignore it.
        file_access (FileAccess):
            This parameter may be used to control how files are opened internally. It is ignored if *input_data* is not a file path.
        autoclose (bool):
            If set to :data:`True` and a byte buffer was provided as input, :meth:`.close` will not only close the PDFium document, but also the input source.
    
    Raises:
        PdfiumError: Raised if the document failed to load. The exception message is annotated with the reason reported by PDFium.
        FileNotFoundError: Raised if an invalid or non-existent file path was given.
    
    Hint:
        * :func:`len` may be called to get a document's number of pages.
        * Looping over a document will yield its pages from beginning to end.
        * Pages may be loaded using list index access.
        * The ``del`` keyword and list index access may be used to delete pages.
    
    Attributes:
        raw (FPDF_DOCUMENT): The underlying PDFium document handle.
    """
    
    def __init__(
            self,
            input_data,
            password = None,
            file_access = FileAccess.NATIVE,
            autoclose = False,
        ):
        
        self._orig_input = input_data
        self._actual_input = input_data
        self._data_holder = []
        self._data_closer = []
        self._rendering_input = None
        
        self._password = password
        self._file_access = file_access
        self._autoclose = autoclose
        
        self._form_env = None
        self._form_config = None
        self._form_finalizer = None
        
        if isinstance(self._orig_input, str):
            
            self._orig_input = os.path.abspath( os.path.expanduser(self._orig_input) )
            if not os.path.isfile(self._orig_input):
                raise FileNotFoundError("File does not exist: '%s'" % self._orig_input)
            
            if self._file_access is FileAccess.NATIVE:
                pass
            elif self._file_access is FileAccess.BUFFER:
                self._actual_input = open(self._orig_input, "rb")
                self._data_closer.append(self._actual_input)
            elif self._file_access is FileAccess.BYTES:
                buf = open(self._orig_input, "rb")
                self._actual_input = buf.read()
                buf.close()
            else:
                assert False
        
        if isinstance(self._actual_input, pdfium.FPDF_DOCUMENT):
            self.raw = self._actual_input
        else:
            self.raw, ld_data = _open_pdf(self._actual_input, self._password)
            self._data_holder += ld_data
        
        if self._autoclose and is_input_buffer(self._actual_input):
            self._data_closer.append(self._actual_input)
        
        self._finalizer = weakref.finalize(
            self, self._static_close,
            self.raw, self._data_holder, self._data_closer,
        )
    
    
    def __enter__(self):
        return self
    
    def __exit__(self, *_):
        # We do not invoke close at this place anymore because that would increase the risk of callers closing parent objects before child objects.
        # (Consider a `with`-block where pages are not closed explicitly: garbage collection of pages commonly happens later than context manager exit, so page would be closed after document, which is illegal.)
        pass
    
    def __len__(self):
        return pdfium.FPDF_GetPageCount(self.raw)
    
    def __iter__(self):
        for i in range( len(self) ):
            yield self.get_page(i)
    
    def __getitem__(self, i):
        return self.get_page(i)
    
    def __delitem__(self, i):
        self.del_page(i)
    
    
    @classmethod
    def new(cls):
        """
        Returns:
            PdfDocument: A new, empty document.
        """
        new_pdf = pdfium.FPDF_CreateNewDocument()
        return cls(new_pdf)
    
    
    @staticmethod
    def _static_close(raw, data_holder, data_closer):
        
        # logger.debug("Closing document")
        pdfium.FPDF_CloseDocument(raw)
        
        for data in data_holder:
            id(data)
        for data in data_closer:
            data.close()
    
    
    @staticmethod
    def _static_exit_formenv(form_env, form_config):
        # logger.debug("Closing form env")
        pdfium.FPDFDOC_ExitFormFillEnvironment(form_env)
        id(form_config)
    
    
    def close(self):
        """
        Free memory by applying the finalizer for the underlying PDFium document.
        Please refer to the generic note on ``close()`` methods for details.
        
        This method calls :meth:`.exit_formenv`.
        """
        if self.raw is None:
            logger.warning("Duplicate close call suppressed on document %s" % self)
            return
        self.exit_formenv()
        self._finalizer()
        self.raw = None
        self._data_holder = []
        self._data_closer = []
    
    
    def _tree_closed(self):
        return (self.raw is None)
    
    
    def init_formenv(self):
        """
        Initialise a form environment handle for this document.
        If already initialised, the existing one will be returned instead.
        
        Returns:
            FPDF_FORMHANDLE:
        """
        if self._form_env is not None:
            return self._form_env
        self._form_config = pdfium.FPDF_FORMFILLINFO()
        self._form_config.version = 2
        self._form_env = pdfium.FPDFDOC_InitFormFillEnvironment(self.raw, self._form_config)
        self._form_finalizer = weakref.finalize(
            self, self._static_exit_formenv,
            self._form_env, self._form_config,
        )
        return self._form_env
    
    
    def exit_formenv(self):
        """
        Free memory by applying the finalizer for the underlying PDFium form environment, if it was initialised.
        If :meth:`.init_formenv` was not called, nothing will be done.
        
        This behaves like the ``close()`` methods. Please refer to the generic note for details.
        """
        if self._form_env is None:
            return
        self._form_finalizer()
        self._form_env = None
        self._form_config = None
    
    
    def get_version(self):
        """
        Returns:
            int | None: The PDF version of the document (14 for 1.4, 15 for 1.5, ...),
            or :data:`None` if the version could not be determined (e. g. because the document was created using :meth:`PdfDocument.new`).
        """
        version = ctypes.c_int()
        success = pdfium.FPDF_GetFileVersion(self.raw, version)
        if not success:
            return
        return int(version.value)
    
    
    def save(self, buffer, version=None):
        """
        Save the document into an output buffer, at its current state.
        
        Parameters:
            buffer (typing.BinaryIO):
                A byte buffer to capture the data.
                It may be any object implementing the ``write()`` method.
            version (int | None):
                 The PDF version to use, given as an integer (14 for 1.4, 15 for 1.5, ...).
                 If :data:`None`, PDFium will set a version automatically.
        """
        
        filewrite = pdfium.FPDF_FILEWRITE()
        filewrite.version = 1
        filewrite.WriteBlock = get_functype(pdfium.FPDF_FILEWRITE, "WriteBlock")( _writer_class(buffer) )
        
        saveargs = (self.raw, filewrite, pdfium.FPDF_NO_INCREMENTAL)
        if version is None:
            success = pdfium.FPDF_SaveAsCopy(*saveargs)
        else:
            success = pdfium.FPDF_SaveWithVersion(*saveargs, version)
        
        if not success:
            raise PdfiumError("Saving the document failed")
    
    
    def _handle_index(self, index):
        n_pages = len(self)
        if index < 0:
            index += n_pages
        if not 0 <= index < n_pages:
            raise IndexError("Page index %s is out of bounds for document with %s pages." % (index, n_pages))
        return index
    
    
    def get_page_size(self, index):
        """
        Get the dimensions of the page at *index*. Reverse indexing is allowed.
        
        Returns:
            (float, float): Page width and height in PDF canvas units.
        """
        index = self._handle_index(index)
        size = pdfium.FS_SIZEF()
        success = pdfium.FPDF_GetPageSizeByIndexF(self.raw, index, size)
        if not success:
            raise PdfiumError("Getting page size by index failed.")
        return (size.width, size.height)
    
    
    def page_as_xobject(self, index, dest_pdf):
        """
        Capture a page as XObject and attach it to a document's resources.
        
        Parameters:
            index (int): Zero-based index of the page. Reverse indexing is allowed.
            dest_pdf (PdfDocument): Target document to which the XObject shall be added.
        Returns:
            PdfXObject: The page as XObject.
        """
        
        index = self._handle_index(index)
        
        raw_xobject = pdfium.FPDF_NewXObjectFromPage(dest_pdf.raw, self.raw, index)
        if raw_xobject is None:
            raise PdfiumError("Failed to capture page %s as FPDF_XOBJECT" % index)
        
        return PdfXObject(
            raw = raw_xobject,
            pdf = dest_pdf,
        )
    
    
    def new_page(self, width, height, index=None):
        """
        Insert a new, empty page into the document.
        
        Parameters:
            width (float):
                Target page width (horizontal size).
            height (float):
                Target page height (vertical size).
            index (int | None):
                Suggested zero-based index at which the page will be inserted.
                If *index* is negative, the indexing direction will be reversed.
                If *index* is zero, the page will be inserted at the beginning.
                If *index* is :data:`None` or larger that the document's current last index, the page will be appended to the end.
        Returns:
            PdfPage: The newly created page.
        """
        if index is None:
            index = len(self)
        elif index < 0:
            index += len(self)
        raw_page = pdfium.FPDFPage_New(self.raw, index, width, height)
        return PdfPage(raw_page, self)
    
    
    def del_page(self, index):
        """
        Remove the page at *index*. Reverse indexing is allowed.
        """
        index = self._handle_index(index)
        pdfium.FPDFPage_Delete(self.raw, index)
    
    
    def get_page(self, index):
        """
        Returns:
            PdfPage: The page at *index*. Reverse indexing is allowed.
        """
        index = self._handle_index(index)
        raw_page = pdfium.FPDF_LoadPage(self.raw, index)
        return PdfPage(raw_page, self)
    
    
    def add_font(self, font_path, type, is_cid):
        """
        Add a font to the document.
        
        Parameters:
            font_path (str):
                File path of the font to use.
            type (int):
                A constant signifying the type of the given font (:data:`.FPDF_FONT_*`).
            is_cid (bool):
                Whether the given font is a CID font or not.
        Returns:
            PdfFont: A PDF font helper object.
        """
        
        with open(font_path, "rb") as fh:
            font_data = fh.read()
        
        pdf_font = pdfium.FPDFText_LoadFont(
            self.raw,
            ctypes.cast(font_data, ctypes.POINTER(ctypes.c_uint8)),
            len(font_data),
            type,
            is_cid,
        )
        
        return PdfFont(pdf_font, self, font_data)
    
    
    def _get_bookmark(self, bookmark, level):
        
        n_bytes = pdfium.FPDFBookmark_GetTitle(bookmark, None, 0)
        buffer = ctypes.create_string_buffer(n_bytes)
        pdfium.FPDFBookmark_GetTitle(bookmark, buffer, n_bytes)
        title = buffer.raw[:n_bytes-2].decode('utf-16-le')
        
        count = pdfium.FPDFBookmark_GetCount(bookmark)
        if count < 0:
            is_closed = True
        elif count == 0:
            is_closed = None
        else:
            is_closed = False
        
        n_kids = abs(count)
        dest = pdfium.FPDFBookmark_GetDest(self.raw, bookmark)
        page_index = pdfium.FPDFDest_GetDestPageIndex(self.raw, dest)
        if page_index == -1:
            page_index = None
        
        n_params = ctypes.c_ulong()
        view_pos = (pdfium.FS_FLOAT * 4)()
        view_mode = pdfium.FPDFDest_GetView(dest, n_params, view_pos)
        view_pos = list(view_pos)[:n_params.value]
        
        return OutlineItem(
            level = level,
            title = title,
            is_closed = is_closed,
            n_kids = n_kids,
            page_index = page_index,
            view_mode = view_mode,
            view_pos = view_pos,
        )
    
    
    def get_toc(
            self,
            max_depth = 15,
            parent = None,
            level = 0,
            seen = None,
        ):
        """
        Read the document's outline ("table of contents").
        
        Parameters:
            max_depth (int):
                Maximum recursion depth to consider when reading the outline.
        Yields:
            :class:`.OutlineItem`: The data of an outline item ("bookmark").
        """
        
        if seen is None:
            seen = set()
        
        bookmark = pdfium.FPDFBookmark_GetFirstChild(self.raw, parent)
        
        while bookmark:
            
            address = ctypes.addressof(bookmark.contents)
            if address in seen:
                logger.warning("A circular bookmark reference was detected whilst parsing the table of contents.")
                break
            else:
                seen.add(address)
            
            yield self._get_bookmark(bookmark, level)
            if level < max_depth-1:
                yield from self.get_toc(
                    max_depth = max_depth,
                    parent = bookmark,
                    level = level + 1,
                    seen = seen,
                )
            
            bookmark = pdfium.FPDFBookmark_GetNextSibling(self.raw, bookmark)
    
    
    @staticmethod
    def print_toc(toc, n_digits=2):
        """
        Print a table of contents.
        
        Parameters:
            toc (typing.Iterator[OutlineItem]):
                Sequence of outline items to show.
            n_digits (int):
                The number of digits to which viewport coordinates shall be rounded.
        """
        
        for item in toc:
            
            if item.n_kids == 0:
                state = "*"
            elif item.is_closed:
                state = "-"
            else:
                state = "+"
            
            if item.page_index is None:
                target = "?"
            else:
                target = item.page_index + 1
            
            print(
                "    " * item.level +
                "[%s] %s -> %s  # %s %s" % (
                    state, item.title, target,
                    ViewmodeToStr[item.view_mode],
                    [round(c, n_digits) for c in item.view_pos],
                )
            )
    
    
    def update_rendering_input(self):
        """
        Update the input sources for concurrent rendering to the document's current state
        by saving to bytes and setting the result as new input.
        If you modified the document, you may want to call this method before :meth:`.render_to`.
        """
        buffer = io.BytesIO()
        self.save(buffer)
        buffer.seek(0)
        self._rendering_input = buffer.read()
        buffer.close()
    
    
    @classmethod
    def _process_page(cls, index, converter, input_data, password, file_access, **kwargs):
        pdf = cls(
            input_data,
            password = password,
            file_access = file_access,
        )
        page = pdf.get_page(index)
        result = page.render_to(converter, **kwargs)
        return result, index
    
    
    def render_to(
            self,
            converter,
            page_indices = None,
            n_processes = os.cpu_count(),
            **kwargs
        ):
        """
        Concurrently render multiple pages, using a process pool executor.
        
        If rendering only a single page, the call is simply forwarded to :meth:`.PdfPage.render_to` as a shortcut.
        
        Parameters:
            page_indices (typing.Sequence[int] | None):
                A sequence of zero-based indices of the pages to render. Reverse indexing or duplicate page indices are prohibited.
                If :data:`None`, all pages will be included. The order of results is guaranteed to match the order of given page indices.
            n_processes (int):
                Target number of parallel processes.
            kwargs (dict):
                Keyword arguments to the renderer. See :meth:`.PdfPage.render_to` / :meth:`.PdfPage.render_base`.
        
        Yields:
            :data:`typing.Any`: Implementation-specific result object.
        """
        
        n_pages = len(self)
        if not page_indices:
            page_indices = [i for i in range(n_pages)]
        else:
            if not all(0 <= i < n_pages for i in page_indices):
                raise ValueError("Out-of-bounds page index")
            if len(page_indices) != len(set(page_indices)):
                raise ValueError("Duplicate page index")
        
        # shortcut: if we're rendering just a single page, don't waste time setting up a process pool
        if len(page_indices) == 1:
            page = self.get_page(page_indices[0])
            result = page.render_to(converter, **kwargs)
            yield result
            return
        
        if self._rendering_input is None:
            if isinstance(self._orig_input, pdfium.FPDF_DOCUMENT):
                logger.warning("Cannot perform concurrent processing without input sources - saving the document implicitly to get picklable data.")
                self.update_rendering_input()
            elif is_input_buffer(self._orig_input):
                logger.warning("Cannot perform concurrent rendering with buffer input - reading the whole buffer into memory implicitly.")
                cursor = self._orig_input.tell()
                self._orig_input.seek(0)
                self._rendering_input = self._orig_input.read()
                self._orig_input.seek(cursor)
            else:
                self._rendering_input = self._orig_input
        
        invoke_renderer = functools.partial(
            PdfDocument._process_page,
            converter = converter,
            input_data = self._rendering_input,
            password = self._password,
            file_access = self._file_access,
            **kwargs
        )
        
        i = 0
        with ProcessPoolExecutor(n_processes) as pool:
            for result, index in pool.map(invoke_renderer, page_indices):
                assert index == page_indices[i]
                i += 1
                yield result
        
        assert len(page_indices) == i


def _open_pdf(input_data, password=None):
    
    if isinstance(password, str):
        password = password.encode("utf-8")
    
    ld_data = ()
    if isinstance(input_data, str):
        pdf = pdfium.FPDF_LoadDocument(input_data.encode("utf-8"), password)
    elif isinstance(input_data, bytes):
        pdf = pdfium.FPDF_LoadMemDocument64(input_data, len(input_data), password)
        ld_data = (input_data, )
    elif is_input_buffer(input_data):
        fileaccess, ld_data = get_fileaccess(input_data)
        pdf = pdfium.FPDF_LoadCustomDocument(fileaccess, password)
    else:
        raise TypeError("Invalid input type '%s'" % type(input_data).__name__)
    
    if pdfium.FPDF_GetPageCount(pdf) < 1:
        err_code = pdfium.FPDF_GetLastError()
        pdfium_msg = ErrorToStr.get(err_code, "Error code %s" % err_code)
        raise PdfiumError("Loading the document failed (PDFium: %s)" % pdfium_msg)
    
    return pdf, ld_data


class _writer_class:
    
    def __init__(self, buffer):
        self.buffer = buffer
        if not callable( getattr(self.buffer, "write", None) ):
            raise ValueError("Output buffer must implement the write() method.")
    
    def __call__(self, _, data, size):
        block = ctypes.cast(data, ctypes.POINTER(ctypes.c_ubyte * size))
        self.buffer.write(block.contents)
        return 1


class PdfXObject:
    """
    XObject helper class.
    
    Attributes:
        raw (FPDF_XOBJECT): The underlying PDFium XObject handle.
        pdf (PdfDocument): Reference to the document this XObject belongs to.
    """
    
    def __init__(self, raw, pdf):
        self.raw = raw
        self.pdf = pdf
        self._finalizer = weakref.finalize(
            self, self._static_close,
            self.raw, self.pdf,
        )
    
    def _tree_closed(self):
        if self.raw is None:
            return True
        return self.pdf._tree_closed()
    
    @staticmethod
    def _static_close(raw, parent):
        # logger.debug("Closing XObject")
        if parent._tree_closed():
            logger.critical("Document closed before XObject (this is illegal). Document: %s" % parent)
        pdfium.FPDF_CloseXObject(raw)
    
    def close(self):
        """
        Free memory by applying the finalizer for the underlying PDFium XObject.
        Please refer to the generic note on ``close()`` methods for details.
        """
        if self.raw is None:
            logger.warning("Duplicate close call suppressed on XObject %s" % self)
            return
        self._finalizer()
        self.raw = None
    
    def as_pageobject(self):
        """
        Returns:
            PdfPageObject: A new pageobject referencing the XObject.
        """
        raw_pageobj = pdfium.FPDF_NewFormObjectFromXObject(self.raw)
        return PdfPageObject(
            raw = raw_pageobj,
            type = pdfium.FPDF_PAGEOBJ_FORM,
            pdf = self.pdf,
        )


class HarfbuzzFont:
    """ Harfbuzz font data helper class. """
    
    def __init__(self, font_path):
        if harfbuzz is None:
            raise RuntimeError("Font helpers require uharfbuzz to be installed.")
        self.blob = harfbuzz.Blob.from_file_path(font_path)
        self.face = harfbuzz.Face(self.blob)
        self.font = harfbuzz.Font(self.face)
        self.scale = self.font.scale[0]


class PdfFont:
    """
    PDF font data helper class.
    
    Attributes:
        raw (FPDF_FONT): The underlying PDFium font handle.
        pdf (PdfDocument): Reference to the document this font belongs to.
    """
    
    def __init__(self, raw, pdf, font_data):
        self.raw = raw
        self.pdf = pdf
        self._font_data = font_data
        self._finalizer = weakref.finalize(
            self, self._static_close,
            self.raw, self.pdf, self._font_data,
        )
    
    def _tree_closed(self):
        if self.raw is None:
            return True
        return self.pdf._tree_closed()
    
    @staticmethod
    def _static_close(raw, parent, font_data):
        # logger.debug("Closing font")
        if parent._tree_closed():
            logger.critical("Document closed before font (this is illegal). Document: %s" % parent)
        pdfium.FPDFFont_Close(raw)
        id(font_data)
    
    def close(self):
        """
        Free memory by applying the finalizer for the underlying PDFium font.
        Please refer to the generic note on ``close()`` methods for details.
        """
        if self.raw is None:
            logger.warning("Duplicate close call suppressed on font %s" % self)
            return
        self._finalizer()
        self.raw = None



# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

# This namespace encompasses all public pypdfium2 members

from pypdfium2._pypdfium import *
from pypdfium2._helpers import *
from pypdfium2.version import *



# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

import os.path
from enum import Enum
from pypdfium2 import _namespace as pdfium


class Units (Enum):
    PT = 0
    MM = 1
    CM = 2
    IN = 3


def units_to_pt(value, unit: Units):
    if unit is Units.PT:
        return value
    elif unit is Units.IN:
        return value*72
    elif unit is Units.CM:
        return (value*72) / 2.54
    elif unit is Units.MM:
        return (value*72) / 25.4
    else:
        raise ValueError("Invalid unit type %s" % unit)


def attach_parser(subparsers):
    parser = subparsers.add_parser(
        "tile",
        help = "Perform page tiling (N-up compositing)",
    )
    parser.add_argument(
        "input",
        help = "PDF file on which to perform N-up compositing",
    )
    parser.add_argument(
        "--password",
        help = "Password to unlock the PDF, if encrypted"
    )
    parser.add_argument(
        "--output", "-o",
        required = True,
        type = os.path.abspath,
        help = "Target path for the new document",
    )
    parser.add_argument(
        "--rows", "-r",
        type = int,
        required = True,
        help = "Number of rows (horizontal tiles)",
    )
    parser.add_argument(
        "--cols", "-c",
        type = int,
        required = True,
        help = "Number of columns (vertical tiles)",
    )
    parser.add_argument(
        "--width",
        type = float,
        required = True,
        help = "Target width",
    )
    parser.add_argument(
        "--height",
        type = float,
        required = True,
        help = "Target height",
    )
    parser.add_argument(
        "--unit", "-u",
        default = Units.MM,
        type = lambda string: Units[string.upper()],
        help = "Unit for target width and height (pt, mm, cm, in)",
    )    


def main(args):
    
    width = units_to_pt(args.width, args.unit)
    height = units_to_pt(args.height, args.unit)
    
    src_pdf = pdfium.PdfDocument(args.input, password=args.password)
    raw_dest = pdfium.FPDF_ImportNPagesToOne(
        src_pdf.raw,
        width, height,
        args.cols, args.rows,
    )
    dest_pdf = pdfium.PdfDocument(raw_dest)
    with open(args.output, "wb") as buffer:
        dest_pdf.save(buffer)



# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

from ctypes import c_float
import pypdfium2._pypdfium as pdfium
from pypdfium2._helpers.misc import (
    PdfiumError,
    get_fileaccess,
    is_input_buffer,
)
from pypdfium2._helpers.matrix import PdfMatrix


class PdfPageObject:
    """
    Page object helper class.
    
    When constructing a :class:`.PdfPageObject`, a more specific subclass may be returned instead,
    depending on :attr:`.type` (e. g. :class:`.PdfImageObject`).
    
    Note:
        Page objects are automatically freed by PDFium with the page they belong to.
        If a page object ends up without associated page, you may want to call ``FPDFPageObj_Destroy(pageobj.raw)``.
    
    Attributes:
        raw (FPDF_PAGEOBJECT):
            The underlying PDFium pageobject handle.
        type (int):
            The type of the object (:data:`FPDF_PAGEOBJ_*`), at the time of initialisation.
        page (PdfPage):
            Reference to the page this pageobject belongs to. May be :data:`None` if the object does not belong to a page yet.
        pdf (PdfDocument):
            Reference to the document this pageobject belongs to. May be :data:`None` if the object does not belong to a document yet.
            This attribute is always set if :attr:`.page` is set.
        level (int):
            Nesting level signifying the number of parent Form XObjects, at the time of initialisation.
            Zero if the object is not nested in a Form XObject.
    """
    
    
    def __new__(cls, raw, type, *args, **kwargs):
        # Allow to create a more specific helper depending on the type
        if type == pdfium.FPDF_PAGEOBJ_IMAGE:
            instance = super().__new__(PdfImageObject)
        else:
            instance = super().__new__(PdfPageObject)
        return instance
    
    
    def __init__(self, raw, type, page=None, pdf=None, level=0):
        
        self.raw = raw
        self.type = type
        self.page = page
        self.pdf = pdf
        self.level = level
        
        if page is not None:
            if self.pdf is None:
                self.pdf = page.pdf
            elif self.pdf is not page.pdf:
                raise ValueError("*page* must belong to *pdf* when constructing a pageobject.")
    
    
    def get_pos(self):
        """
        Get the position of the object on the page.
        
        Returns:
            A tuple of four :class:`float` coordinates for left, bottom, right, and top.
        """
        if self.page is None:
            raise RuntimeError("Must not call get_pos() on a loose pageobject.")
        
        left, bottom, right, top = c_float(), c_float(), c_float(), c_float()
        success = pdfium.FPDFPageObj_GetBounds(self.raw, left, bottom, right, top)
        if not success:
            raise PdfiumError("Failed to locate pageobject.")
        
        return (left.value, bottom.value, right.value, top.value)
    
    
    def get_matrix(self):
        """
        Returns:
            PdfMatrix: The pageobject's current transform matrix.
        """
        fs_matrix = pdfium.FS_MATRIX()
        success = pdfium.FPDFPageObj_GetMatrix(self.raw, fs_matrix)
        if not success:
            raise PdfiumError("Failed to get matrix of pageobject.")
        return PdfMatrix.from_pdfium(fs_matrix)
    
    
    def set_matrix(self, matrix):
        """
        Define *matrix* as the pageobject's transform matrix.
        """
        if not isinstance(matrix, PdfMatrix):
            raise ValueError("*matrix* must be a PdfMatrix object.")
        success = pdfium.FPDFPageObj_SetMatrix(self.raw, matrix.to_pdfium())
        if not success:
            raise PdfiumError("Failed to set matrix of pageobject.")
    
    
    def transform(self, matrix):
        """
        Apply *matrix* on top of the pageobject's current transform matrix.
        """
        if not isinstance(matrix, PdfMatrix):
            raise ValueError("*matrix* must be a PdfMatrix object.")
        pdfium.FPDFPageObj_Transform(self.raw, *matrix.get())


class PdfImageObject (PdfPageObject):
    """
    Image object helper class (specific kind of page object).
    """
    
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
    
    @classmethod
    def new(cls, pdf):
        """
        Parameters:
            pdf (PdfDocument): The document to which the new image object shall be added.
        Returns:
            PdfImageObject: Handle to a new, empty image.
        """
        return cls(
            pdfium.FPDFPageObj_NewImageObj(pdf.raw),
            pdfium.FPDF_PAGEOBJ_IMAGE,
            page = None,
            pdf = pdf,
        )
    
    
    def load_jpeg(self, buffer, pages=None, inline=False, autoclose=True):
        """
        Load a JPEG into the image object.
        
        Position and size of the image are defined by its matrix.
        If the image is new, it will appear as a tiny square of 1x1 units on the bottom left corner of the page.
        Use :class:`.PdfMatrix` and :meth:`.set_matrix` to adjust the position.
        
        If replacing an image, the existing matrix will be preserved.
        If aspect ratios do not match, the new image will be squashed into the old image's boundaries.
        Modify the matrix manually if you wish to prevent this.
        
        Parameters:
            buffer (typing.BinaryIO):
                A readable byte buffer to access the JPEG data.
            pages (typing.Sequence[PdfPage] | None):
                If replacing an image, pass in a list of loaded pages that might contain the it, to update their cache.
                (The same image may be shown multiple times in different transforms across a PDF.)
                If the image object handle is new, this parameter may be :data:`None` or an empty list.
            inline (bool):
                Whether to load the image content into memory.
                If :data:`True`, the buffer may be closed after this function call.
                Otherwise, the buffer needs to remain open until the PDF is closed.
            autoclose (bool):
                Whether the buffer should be automatically closed once it is not needed anymore.
        
        Returns:
            (int, int): Image width and height in pixels.
        """
        
        if not is_input_buffer(buffer):
            raise ValueError("This is not a compatible buffer: %s" % buffer)
        
        fileaccess, ld_data = get_fileaccess(buffer)
        
        if inline:
            loader = pdfium.FPDFImageObj_LoadJpegFileInline
        else:
            loader = pdfium.FPDFImageObj_LoadJpegFile
        
        c_pages = None
        page_count = 0
        if pages:
            page_count = len(pages)
            c_pages = (pdfium.FPDF_PAGE * page_count)(*[p.raw for p in pages])
        
        success = loader(c_pages, page_count, self.raw, fileaccess)
        if not success:
            raise PdfiumError("Loading JPEG into image object failed.")
        
        if inline:
            id(ld_data)
            if autoclose:
                buffer.close()
        else:
            self.pdf._data_holder += ld_data
            if autoclose:
                self.pdf._data_closer.append(buffer)
        
        metadata = self.get_info()
        return (metadata.width, metadata.height)
    
    
    def get_info(self):
        """
        Returns:
            FPDF_IMAGEOBJ_METADATA:
            A structure containing information about the image object, including dimensions, DPI, bits per pixel, and colour space.
            If the image does not belong to a page yet, some values will be unset (0).
        """
        
        raw_page = (self.page.raw if self.page else None)
        
        metadata = pdfium.FPDF_IMAGEOBJ_METADATA()
        success = pdfium.FPDFImageObj_GetImageMetadata(self.raw, raw_page, metadata)
        if not success:
            raise PdfiumError("Failed to retrieve image metadata.")
        
        return metadata



# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

import os.path
from pypdfium2 import _namespace as pdfium


def attach_parser(subparsers):
    parser = subparsers.add_parser(
        "jpegtopdf",
        help = "Convert JPEG images to PDF",
    )
    parser.add_argument(
        "images",
        nargs = "+",
        help = "Input JPEG images",
        type = os.path.abspath,
    )
    parser.add_argument(
        "--output", "-o",
        required = True,
        help = "Target path for the new PDF"
    )
    parser.add_argument(
        "--inline",
        action = "store_true",
        help = "Whether to use FPDFImageObj_LoadJpegFileInline() rather than FPDFImageObj_LoadJpegFile()."
    )


def main(args):
    
    # Very rudimentary JPEG to PDF conversion, mostly for testing
    # The implementation could certainly be more sophisticated (e. g. configurable DPI, default DPI based on image metadata via Pillow, margins, crop, positioning, ...)
    
    pdf = pdfium.PdfDocument.new()
    
    for file in args.images:
        
        # Simple check if the input files are actually JPEGs
        # A better implementation could use mimetypes or python-magic instead
        assert any(file.lower().endswith(ext) for ext in (".jpg", ".jpeg"))
        
        image = pdfium.PdfImageObject.new(pdf)
        
        buffer = open(file, "rb")
        width, height = image.load_jpeg(buffer, inline=args.inline, autoclose=True)
        
        page = pdf.new_page(width, height)
        page.insert_object(image)
        page.generate_content()
    
    if os.path.exists(args.output):
        raise FileExistsError("Refusing to overwrite '%s'" % args.output)
    
    with open(args.output, "wb") as buffer:
        pdf.save(buffer)



# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

import os
import sys
import shutil
import argparse
import tempfile
from os.path import (
    join,
    dirname,
    abspath,
)

sys.path.insert(0, dirname(dirname(abspath(__file__))))
from pl_setup.packaging_base import (
    run_cmd,
    clean_artefacts,
    Host,
    ModuleDir,
    BindingsFileName,
    LibnameForSystem,
    BinaryPlatforms,
    SourceTree,
    BinaryTargetVar,
    BinaryTarget_None,
)


class ArtefactStash:
    
    # Preserve in-tree aftefacts from editable install
    
    def __init__(self):
        
        self.tmp_dir = None
        self.plfile_names = []
        self.plfile_paths = []
        
        patterns = (BindingsFileName, LibnameForSystem[Host.system])
        for fn in patterns:
            fp = join(ModuleDir, fn)
            if not os.path.exists(fp):
                continue
            self.plfile_names.append(fn)
            self.plfile_paths.append(fp)
        assert len(self.plfile_names) == len(self.plfile_paths)
        
        if len(self.plfile_paths) == 0:
            return
        elif len(self.plfile_paths) != 2:
            print("Warning: Expected exactly 2 platform files, but found %s." % len(self.plfile_paths), file=sys.stderr)
        
        self.tmp_dir = tempfile.TemporaryDirectory(prefix="pypdfium2_artefact_stash_")
        for fp in self.plfile_paths:
            shutil.move(fp, self.tmp_dir.name)
    
    def pop(self):
        if self.tmp_dir is None:
            return
        for fn in self.plfile_names:
            shutil.move(join(self.tmp_dir.name, fn), ModuleDir)
        self.tmp_dir.cleanup()


def run_build(args):
    run_cmd([sys.executable, "-m", "build", "--skip-dependency-check", "--no-isolation"] + args, cwd=SourceTree, env=os.environ)


def main():
    
    parser = argparse.ArgumentParser(
        description = "Craft sdist and wheels for pypdfium2, using `python3 -m build`. (This script does not take any arguments.)",
    )
    args = parser.parse_args()
    
    stash = ArtefactStash()
    
    os.environ[BinaryTargetVar] = BinaryTarget_None
    run_build(["--sdist"])
    clean_artefacts()
    
    for plat in BinaryPlatforms:
        os.environ[BinaryTargetVar] = plat
        run_build(["--wheel"])
        clean_artefacts()
    
    stash.pop()


if __name__ == '__main__':
    main()



# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

import os.path
from pypdfium2 import _namespace as pdfium
from pypdfium2._cli._parsers import pagetext_type


def attach_parser(subparsers):
    obj_types = list(pdfium.ObjectTypeToConst.keys())
    parser = subparsers.add_parser(
        "find-pageobjects",
        help = "Locate page objects of given types.",
    )
    parser.add_argument(
        "input",
        type = os.path.abspath,
        help = "Path to the PDF document to work with.",
    )
    parser.add_argument(
        "--password",
        help = "Password to unlock the PDF, if encrypted."
    )
    parser.add_argument(
        "--pages",
        type = pagetext_type,
        help = "The pages to search (defaults to all).",
    )
    parser.add_argument(
        "--types",
        nargs = "+",
        metavar = "T",
        choices = obj_types,
        default = obj_types,
        help = "Object types to consider (defaults to all). Choices: %s" % obj_types,
    )
    parser.add_argument(
        "--max-depth",
        type = int,
        default = 2,
        help = "Maximum recursion depth to consider when descending into Form XObjects.",
    )


def main(args):
    
    doc = pdfium.PdfDocument(args.input, password=args.password)
    args.types = [pdfium.ObjectTypeToConst[t] for t in args.types]
    if args.pages is None:
        args.pages = [i for i in range(len(doc))]
    
    for index in args.pages:
        page = doc.get_page(index)
        for obj in page.get_objects(max_depth=args.max_depth):
            if obj.type in args.types:
                print("    "*obj.level + pdfium.ObjectTypeToStr[obj.type], obj.get_pos())



# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

# No external dependencies shall be imported in this file

import os
import shutil
import platform
import sysconfig
import subprocess
from glob import glob
from os.path import (
    join,
    abspath,
    dirname,
    expanduser,
)

# TODO improve consistency of variable names; think about variables to move in/out

BinaryTargetVar   = "PDFIUM_BINARY"
BinaryTarget_None = "none"
BinaryTarget_Auto = "auto"
BindingsFileName  = "_pypdfium.py"
VerStatusFileName = ".pdfium_version.txt"
HomeDir     = expanduser("~")
SourceTree  = dirname(dirname(dirname(abspath(__file__))))
DataTree    = join(SourceTree, "data")
SB_Dir      = join(SourceTree, "sourcebuild")
ModuleDir   = join(SourceTree, "src", "pypdfium2")
VersionFile = join(ModuleDir, "version.py")
Changelog   = join(SourceTree, "docs", "devel", "changelog.md")
ChangelogStaging = join(SourceTree, "docs", "devel", "changelog_staging.md")
RepositoryURL  = "https://github.com/pypdfium2-team/pypdfium2"
PDFium_URL     = "https://pdfium.googlesource.com/pdfium"
DepotTools_URL = "https://chromium.googlesource.com/chromium/tools/depot_tools.git"
ReleaseRepo    = "https://github.com/bblanchon/pdfium-binaries"
ReleaseURL     = ReleaseRepo + "/releases/download/chromium%2F"


class SystemNames:
    linux   = "linux"
    darwin  = "darwin"
    windows = "windows"


class PlatformNames:
    # - Attribute names and values are expected to match
    # - Platform names are expected to start with the corresponding system name
    linux_x64      = SystemNames.linux   + "_x64"
    linux_x86      = SystemNames.linux   + "_x86"
    linux_arm64    = SystemNames.linux   + "_arm64"
    linux_arm32    = SystemNames.linux   + "_arm32"
    linux_musl_x64 = SystemNames.linux   + "_musl_x64"
    linux_musl_x86 = SystemNames.linux   + "_musl_x86"
    darwin_x64     = SystemNames.darwin  + "_x64"
    darwin_arm64   = SystemNames.darwin  + "_arm64"
    windows_x64    = SystemNames.windows + "_x64"
    windows_x86    = SystemNames.windows + "_x86"
    windows_arm64  = SystemNames.windows + "_arm64"
    sourcebuild    = "sourcebuild"


ReleaseNames = {
    PlatformNames.darwin_x64     : "pdfium-mac-x64",
    PlatformNames.darwin_arm64   : "pdfium-mac-arm64",
    PlatformNames.linux_x64      : "pdfium-linux-x64",
    PlatformNames.linux_x86      : "pdfium-linux-x86",
    PlatformNames.linux_arm64    : "pdfium-linux-arm64",
    PlatformNames.linux_arm32    : "pdfium-linux-arm",
    PlatformNames.linux_musl_x64 : "pdfium-linux-musl-x64",
    PlatformNames.linux_musl_x86 : "pdfium-linux-musl-x86",
    PlatformNames.windows_x64    : "pdfium-win-x64",
    PlatformNames.windows_x86    : "pdfium-win-x86",
    PlatformNames.windows_arm64  : "pdfium-win-arm64",
}

LibnameForSystem = {
    SystemNames.linux:   "pdfium",
    SystemNames.darwin:  "pdfium.dylib",
    SystemNames.windows: "pdfium.dll",
}

BinaryPlatforms = list(ReleaseNames.keys())
BinarySystems   = list(LibnameForSystem.keys())
MainLibnames    = list(LibnameForSystem.values())

def plat_to_system(pl_name):
    if pl_name == PlatformNames.sourcebuild:
        # NOTE If doing a sourcebuild on an unknown host system, this returns None, which will cause binary detection code to fail (we need to know the platform-specific binary name).
        return Host.system
    result = [s for s in BinarySystems if pl_name.startswith(s)]
    assert len(result) == 1
    return result[0]


class _host_platform:
    
    def __init__(self):
        
        # `libc_ver()` currently returns an empty string on libc implementations other than glibc - hence, we assume musl if it's not glibc
        # FIXME is there some function to actually detect musl?
        self._plat_info = sysconfig.get_platform().lower().replace("-", "_").replace(".", "_")
        self._libc_info, self._is_glibc = None, None
        if self._plat_info.startswith("linux"):
            self._libc_info = platform.libc_ver()
            self._is_glibc = (self._libc_info[0] == "glibc")
        
        self.platform = self._get_platform()
        self.system = None
        if self.platform is not None:
            self.system = plat_to_system(self.platform)
    
    def _is_plat(self, start, end):
        return self._plat_info.startswith(start) and self._plat_info.endswith(end)
    
    def _get_platform(self):
        if self._is_plat("macosx", "arm64"):
            return PlatformNames.darwin_arm64
        elif self._is_plat("macosx", "x86_64"):
            return PlatformNames.darwin_x64
        elif self._is_plat("linux", "armv7l"):
            return PlatformNames.linux_arm32
        elif self._is_plat("linux", "aarch64"):
            return PlatformNames.linux_arm64
        elif self._is_plat("linux", "x86_64"):
            return PlatformNames.linux_x64 if self._is_glibc else PlatformNames.linux_musl_x64
        elif self._is_plat("linux", "i686"):
            return PlatformNames.linux_x86 if self._is_glibc else PlatformNames.linux_musl_x86
        elif self._is_plat("win", "arm64"):
            return PlatformNames.windows_arm64
        elif self._is_plat("win", "amd64"):
            return PlatformNames.windows_x64
        elif self._is_plat("win32", ""):
            return PlatformNames.windows_x86
        else:
            return None


Host = _host_platform()


def _get_linux_tag(arch):
    return "manylinux_2_17_%s.manylinux2014_%s" % (arch, arch)

def _get_musllinux_tag(arch):
    return "musllinux_1_2_%s" % (arch)


def _get_mac_tag(arch, *versions):
    
    assert len(versions) > 0
    
    template = "macosx_%s_%s"
    
    tag = ""
    sep = ""
    for v in versions:
        tag += sep + template % (v, arch)
        sep = "."
    
    return tag


def get_wheel_tag(pl_name):
    # pip>=20.3 now accepts macOS wheels tagged as 10_x on 11_x. Not sure what applies to 12_x.
    # Let's retain multi-version tagging for broader compatibility all the same.
    if pl_name == PlatformNames.darwin_x64:
        # pdfium-binaries/steps/05-configure.sh defines `mac_deployment_target = "10.13.0"`
        return _get_mac_tag("x86_64", "10_13", "11_0", "12_0")
    elif pl_name == PlatformNames.darwin_arm64:
        return _get_mac_tag("arm64", "11_0", "12_0")
    elif pl_name == PlatformNames.linux_x64:
        return _get_linux_tag("x86_64")
    elif pl_name == PlatformNames.linux_x86:
        return _get_linux_tag("i686")
    elif pl_name == PlatformNames.linux_arm64:
        return _get_linux_tag("aarch64")
    elif pl_name == PlatformNames.linux_arm32:
        return _get_linux_tag("armv7l")
    elif pl_name == PlatformNames.linux_musl_x64:
        return _get_musllinux_tag("x86_64")
    elif pl_name == PlatformNames.linux_musl_x86:
        return _get_musllinux_tag("i686")
    elif pl_name == PlatformNames.windows_x64:
        return "win_amd64"
    elif pl_name == PlatformNames.windows_arm64:
        return "win_arm64"
    elif pl_name == PlatformNames.windows_x86:
        return "win32"
    elif pl_name == PlatformNames.sourcebuild:
        tag = sysconfig.get_platform()
        for char in ("-", "."):
            tag = tag.replace(char, "_")
        return tag
    else:
        raise ValueError("Unknown platform name %s" % pl_name)


def run_cmd(command, cwd, capture=False, **kwargs):
    
    print('%s ("%s")' % (command, cwd))
    if capture:
        kwargs.update( dict(stdout=subprocess.PIPE, stderr=subprocess.STDOUT) )
    
    comp_process = subprocess.run(command, cwd=cwd, **kwargs)
    if capture:
        return comp_process.stdout.decode("utf-8").strip()
    else:
        return comp_process


def get_latest_version():
    git_ls = run_cmd(["git", "ls-remote", "%s.git" % ReleaseRepo], cwd=None, capture=True)
    tag = git_ls.split("\t")[-1]
    return int( tag.split("/")[-1] )


def call_ctypesgen(target_dir, include_dir):
    
    bindings = join(target_dir, BindingsFileName)
    
    ctypesgen_cmd = ["ctypesgen", "--library", "pdfium", "--strip-build-path", target_dir, "-L", "."] + sorted(glob( join(include_dir, "*.h") )) + ["-o", bindings]
    run_cmd(ctypesgen_cmd, cwd=target_dir)
    
    with open(bindings, "r") as file_reader:
        text = file_reader.read()
        text = text.replace(target_dir, ".")
        text = text.replace(HomeDir, "~")
    
    with open(bindings, "w") as file_writer:
        file_writer.write(text)


def clean_artefacts():
    
    deletables = [
        join(SourceTree, "build"),
        join(ModuleDir, BindingsFileName),
    ]
    deletables += [join(ModuleDir, fn) for fn in MainLibnames]
    
    for item in deletables:
        if not os.path.exists(item):
            continue
        if os.path.isfile(item):
            os.remove(item)
        elif os.path.isdir(item):
            shutil.rmtree(item)


def get_platfiles(pl_name):
    system = plat_to_system(pl_name)
    platfiles = (
        join(DataTree, pl_name, BindingsFileName),
        join(DataTree, pl_name, LibnameForSystem[system])
    )
    return platfiles


def copy_platfiles(pl_name):
    platfiles = get_platfiles(pl_name)
    for fp in platfiles:
        if not os.path.exists(fp):
            raise RuntimeError("Platform file missing: %s" % fp)
        shutil.copy(fp, ModuleDir)


def get_changelog_staging(flush=False):
    
    with open(ChangelogStaging, "r") as fh:
        content = fh.read()
        pos = content.index("\n", content.index("# Changelog")) + 1
        header = content[:pos].strip() + "\n"
        devel_msg = content[pos:].strip()
        if devel_msg:
            devel_msg += "\n"
    
    if flush:
        with open(ChangelogStaging, "w") as fh:
            fh.write(header)
    
    return devel_msg

def get_version_ns():
    ver_ns = {}
    with open(VersionFile, "r") as fh:
        exec(fh.read(), ver_ns)
    ver_ns = {k: v for k, v in ver_ns.items() if not k.startswith("_")}
    return ver_ns

VerNamespace = get_version_ns()


def set_versions(ver_changes):
    
    if len(ver_changes) == 0:
        return False
    
    skip = {var for var, value in ver_changes.items() if value == VerNamespace[var]}
    if len(skip) == len(ver_changes):
        return False
    
    with open(VersionFile, "r") as fh:
        content = fh.read()
    
    for var, new_val in ver_changes.items():
        
        if var in skip:
            continue
        
        # this does not work universally - only one notation per type is supported, and switches between str and non-str types don't work
        # FIXME see if we can restructure this code for improved flexibility
        if isinstance(new_val, str):
            template = '%s = "%s"'
        else:
            template = '%s = %s'
        previous = template % (var, VerNamespace[var])
        updated = template % (var, new_val)
        
        print("'%s' -> '%s'" % (previous, updated))
        assert content.count(previous) == 1
        content = content.replace(previous, updated)
        
        # Beware: While this updates the VerNamespace entry itself, it will not update dependent entries, which may lead to inconsistent data. That is, no reliance can be placed upon the values of dynamic variables (V_PYPDFIUM2 !) after this method has been run. If you need the real value, VerNamespace needs to be re-created.
        VerNamespace[var] = new_val
    
    with open(VersionFile, "w") as fh:
        fh.write(content)
    
    return True



#! /usr/bin/env python3
# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

import os
import sys
import setuptools
from os.path import (
    join,
    abspath,
    dirname,
    exists,
)

sys.path.insert(0, join(dirname(abspath(__file__)), "setupsrc"))
from pl_setup import check_deps
from pl_setup.packaging_base import (
    Host,
    DataTree,
    BinaryTargetVar,
    BinaryTarget_None,
    VerStatusFileName,
    PlatformNames,
    get_platfiles,
    get_latest_version,
)

# NOTE Setuptools may, unfortunately, run this code several times (if using PEP 517 style setup).

LockFile = join(DataTree, ".lock_autoupdate.txt")


def install_handler():
    
    from pl_setup import update_pdfium
    from pl_setup.setup_base import mkwheel
    
    pl_name = Host.platform
    if pl_name is None:
        # If PDFium had a proper build system, we could trigger a source build here
        raise RuntimeError(
            "No pre-built binaries available for platform '%s' with libc implementation '%s'. " % (Host._plat_info, Host._libc_info) +
            "You can attempt a source build, but it's unlikely to work out due to binary toolchain requirements of PDFium's build system. Doing cross-compilation or using a different build system might be possible, though. Please get in touch with the project maintainers."
        )
    
    
    need_update = False
    pl_dir = join(DataTree, pl_name)
    ver_file = join(pl_dir, VerStatusFileName)
    
    if not os.path.exists(pl_dir):
        need_update = True  # platform directory doesn't exist yet
    elif not os.path.exists(ver_file) or not all(exists(fp) for fp in get_platfiles(pl_name)):
        print("Warning: Specific platform files are missing -> implicit update", file=sys.stderr)
        need_update = True
    
    elif not exists(LockFile):
        
        # Automatic updates imply some duplication across different runs. The code runs quickly enough, so this is not much of a problem.
        
        latest_ver = get_latest_version()
        with open(ver_file, "r") as fh:
            curr_version = int( fh.read().strip() )
        
        if curr_version > latest_ver:
            raise RuntimeError("Current version must not be greater than latest")
        if curr_version < latest_ver:
            need_update = True
    
    if need_update:
        update_pdfium.main([pl_name])
    mkwheel(pl_name)


def packaging_handler(target):
    
    from pl_setup.setup_base import mkwheel, SetupKws
    
    if target == BinaryTarget_None:
        setuptools.setup(**SetupKws)
    elif hasattr(PlatformNames, target):
        mkwheel( getattr(PlatformNames, target) )
    else:
        raise ValueError("Invalid deployment target '%s'" % target)
    
    return False


def main():
    
    target = os.environ.get(BinaryTargetVar, None)
    
    if target in (None, "auto"):
        # As check_deps should only need to be run once, we could prevent repeated runs using a status file. However, it runs quickly enough, so this isn't necessary.
        check_deps.main()
        install_handler()
    else:
        packaging_handler(target)


if __name__ == "__main__":
    main()



# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

__all__ = ["V_PYPDFIUM2", "V_LIBPDFIUM", "IS_SOURCEBUILD"]

V_MAJOR = 3
V_MINOR = 6
V_PATCH = 0
V_BETA = None

#: pypdfium2 version string.
V_PYPDFIUM2 = "%s.%s.%s" % (V_MAJOR, V_MINOR, V_PATCH)
if V_BETA is not None:
    V_PYPDFIUM2 += "b%s" % V_BETA

#: PDFium library version string (git tag or commit hash).
V_LIBPDFIUM = "5392"

#: Whether the included PDFium binary was built from source locally (:data:`True`) or downloaded from pdfium-binaries (:data:`False`).
IS_SOURCEBUILD = False



# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

import sys
import setuptools
from os.path import (
    join,
    exists,
    abspath,
    dirname,
)
from wheel.bdist_wheel import bdist_wheel

sys.path.insert(0, dirname(dirname(abspath(__file__))))
from pl_setup.packaging_base import (
    DataTree,
    VerNamespace,
    LibnameForSystem,
    VerStatusFileName,
    PlatformNames,
    plat_to_system,
    get_wheel_tag,
    clean_artefacts,
    copy_platfiles,
    set_versions,
)


def bdist_factory(pl_name):
    
    class pypdfium_bdist (bdist_wheel):
        
        def finalize_options(self, *args, **kws):
            bdist_wheel.finalize_options(self, *args, **kws)
            self.root_is_pure = False
        
        def get_tag(self, *args, **kws):
            return "py3", "none", get_wheel_tag(pl_name)
    
    return pypdfium_bdist


class BinaryDistribution (setuptools.Distribution):
    def has_ext_modules(self):
        return True


SetupKws = dict(
    version = VerNamespace["V_PYPDFIUM2"],
)


def mkwheel(pl_name):
    
    system = plat_to_system(pl_name)
    libname = LibnameForSystem[system]
    
    pl_dir = join(DataTree, pl_name)
    if not exists(pl_dir):
        raise RuntimeError("Missing platform directory %s - you might have forgotten to run update_pdfium.py" % pl_name)
    
    ver_file = join(pl_dir, VerStatusFileName)
    if not exists(ver_file):
        raise RuntimeError("Missing PDFium version file for %s" % pl_name)
    
    with open(ver_file, "r") as fh:
        v_libpdfium = fh.read().strip()
    
    ver_changes = dict()
    ver_changes["V_LIBPDFIUM"] = str(v_libpdfium)
    ver_changes["IS_SOURCEBUILD"] = (pl_name == PlatformNames.sourcebuild)
    set_versions(ver_changes)
    
    clean_artefacts()
    copy_platfiles(pl_name)
    
    setuptools.setup(
        package_data = {"": [libname]},
        cmdclass = {"bdist_wheel": bdist_factory(pl_name)},
        distclass = BinaryDistribution,
        **SetupKws,
    )



# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

from pypdfium2._helpers.misc import BitmapStrReverseToRegular

try:
    import PIL.Image
except ImportError:
    PIL = None

try:
    import numpy.ctypeslib
except ImportError:
    numpy = None


class BitmapConvBase:
    """
    Parent class for bitmap converters compatible with :meth:`.PdfPage.render_to` / :meth:`.PdfDocument.render_to`.
    The initialiser captures any arguments and adds them to the :meth:`.run` call.
    """
    
    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs
    
    @staticmethod
    def run(result, renderer_kws, *args, **kwargs):
        """
        The actual converter function, to be implemented by the inheriting class.
        
        Parameters:
            result (tuple):
                Result of the :meth:`~.PdfPage.render_base` call (ctypes array, colour format, size).
            renderer_kws (dict):
                Dictionary of rendering keywords that were passed in by the caller.
            args (tuple):
                Further positional arguments to the converter, as captured by the initialiser.
            kwargs (dict):
                Further keyword arguments to the converter, as captured by the initialiser.
        Returns:
            typing.Any: The converted rendering result (implementation-specific).
        """
        raise NotImplementedError("Inheriting class must provide run() method.")


class BitmapConv:
    """
    Built-in converters to be applied on the rendering result.
    """
    
    class any (BitmapConvBase):
        """
        Simple factory for converters that merely work with the ctypes array, while passing through additional information unaffected.
        
        Example:
            ``render_to(BitmapConv.any(bytes), **kwargs)``:
                Get an independent copy of the pixel data as bytes.
        
        Parameters:
            converter (typing.Callable):
                A function to translate a ctypes array to a different data type.
        Returns:
            (typing.Any, str, (int, int)): The converted bitmap (implementation-specific), and additional information returned by :meth:`~.PdfPage.render_base` (colour format, size).
        """
        
        @staticmethod
        def run(result, renderer_kws, converter):
            c_array, cl_format, size = result
            return converter(c_array), cl_format, size
    
    
    class numpy_ndarray (BitmapConvBase):
        """
        *Requires* :mod:`numpy`
        
        Get the bitmap as shaped NumPy array referencing the original ctypes array.
        This converter never makes a copy of the data.
        
        Returns:
            (numpy.ndarray, str): NumPy array, and colour format.
        """
        
        @staticmethod
        def run(result, renderer_kws):
            
            if numpy is None:
                raise RuntimeError("NumPy library needs to be installed for numpy_ndarray() converter.")
            
            c_array, cl_format, (width, height) = result
            np_array = numpy.ctypeslib.as_array(c_array)
            np_array.shape = (height, width, len(cl_format))
            
            return np_array, cl_format
    
    
    class pil_image (BitmapConvBase):
        """
        *Requires* :mod:`PIL`
        
        Get the bitmap as PIL image.
        
        Parameters:
            prefer_la (bool):
                If :data:`True`, automatically convert ``RGBA``/``BGRA`` to ``LA`` if rendering in greyscale mode with alpha channel
                (PDFium does not provide ``LA`` output directly).
        Returns:
            PIL.Image.Image: The image object.
        
        Hint:
            This uses :func:`PIL.Image.frombuffer` under the hood.
            If possible for the colour format in question, the image will reference the ctypes array. Otherwise, PIL may create a copy of the data.
            Among the pixel formats supported by PDFium, PIL can directly work with ``RGBA``, ``RGBX`` or ``L``.
            You may want to consider setting the rendering parameters *rev_byteorder* and *prefer_bgrx* to :data:`True` to generate natively compatible output.
        """
        
        @staticmethod
        def run(result, renderer_kws, prefer_la=False):
            
            if PIL is None:
                raise RuntimeError("Pillow library needs to be installed for pil_image() converter.")
            
            c_array, cl_src, size = result
            cl_dst = cl_src
            if cl_src in BitmapStrReverseToRegular.keys():
                cl_dst = BitmapStrReverseToRegular[cl_src]
            
            pil_image = PIL.Image.frombuffer(cl_dst, size, c_array, "raw", cl_src, 0, 1)
            if prefer_la:
                if renderer_kws.get("greyscale", False) and cl_dst == "RGBA":
                    pil_image = pil_image.convert("LA")
            
            return pil_image


class BitmapConvAliases:
    """
    Base class containing rendering target aliases.
    Currently retained for backwards compatibility, but may be deprecated in the future.
    Consider using the :meth:`.PdfPage.render_to` / :meth:`.PdfDocument.render_to` APIs instead.
    """
    
    def render_to(self):
        """ Method to be implemented by the inheriting class. """
        raise NotImplementedError("Inheriting class must provide render_to() method.")
    
    def render_tobytes(self, **kwargs):
        """ Alias for ``render_to(BitmapConv.any(bytes), ...)``. """
        return self.render_to(BitmapConv.any(bytes), **kwargs)
    
    def render_tonumpy(self, **kwargs):
        """ Alias for ``render_to(BitmapConv.numpy_ndarray, ...)``. """
        return self.render_to(BitmapConv.numpy_ndarray, **kwargs)
    
    def render_topil(self, prefer_la=False, **kwargs):
        """ Alias for ``render_to(BitmapConv.pil_image, ...)``. """
        return self.render_to(BitmapConv.pil_image(prefer_la=prefer_la), **kwargs)



# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

from pypdfium2._helpers.misc import *
from pypdfium2._helpers.matrix import *
from pypdfium2._helpers.converters import *
from pypdfium2._helpers.document import *
from pypdfium2._helpers.page import *
from pypdfium2._helpers.pageobject import *
from pypdfium2._helpers.textpage import *



#! /usr/bin/env python3
# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

import shutil


Commands = (
    "git",
    "gcc",
    "ctypesgen",
)

def main():
    missing = {cmd for cmd in Commands if not shutil.which(cmd)}
    if len(missing) > 0:
        raise RuntimeError("The following packages or commands are missing: %s" % missing)


if __name__ == "__main__":
    main()



# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

from pypdfium2 import _namespace as pdfium


def _merge_files(input_paths, passwords):
    
    dest_pdf = pdfium.PdfDocument.new()
    index = 0
    
    for in_path, pwd in zip(input_paths, passwords):
        
        src_pdf = pdfium.PdfDocument(in_path, password=pwd)
        success = pdfium.FPDF_ImportPagesByIndex(dest_pdf.raw, src_pdf.raw, None, 0, index)
        if not success:
            raise RuntimeError("Importing pages failed.")
        
        index += len(src_pdf)
    
    return dest_pdf


def attach_parser(subparsers):
    parser = subparsers.add_parser(
        "merge",
        help = "Concatenate PDF files",
    )
    parser.add_argument(
        "inputs",
        nargs = "+",
        help = "A sequence of PDF files to concatenate",
    )
    parser.add_argument(
        "--passwords",
        nargs = "*",
        help = "A sequence of passwords to unlock encrypted PDFs. The value is ignored for non-encrypted documents, where any placeholder may be used.",
    )
    parser.add_argument(
        "--output", "-o",
        required = True,
        help = "Target path for the output document",
    )


def main(args):
    
    if not args.passwords:
        args.passwords = [None for _ in args.inputs]
    
    merged_pdf = _merge_files(args.inputs, args.passwords)
    with open(args.output, "wb") as buffer:
        merged_pdf.save(buffer)



# PYTHON_ARGCOMPLETE_OK
# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

import sys
import logging
import argparse
from pypdfium2.version import (
    V_PYPDFIUM2,
    V_LIBPDFIUM,
    IS_SOURCEBUILD,
)
from pypdfium2._cli import (
    render,
    toc,
    merge,
    tile,
    extract_text,
    find_pageobjects,
    jpegtopdf,
)

try:
    import argcomplete
except ImportError:
    argcomplete = None


Subcommands = {
    "render": render,
    "toc": toc,
    "merge": merge,
    "tile": tile,
    "extract-text": extract_text,
    "find-pageobjects": find_pageobjects,
    "jpegtopdf": jpegtopdf,
}


def parse_args(argv=sys.argv[1:]):
    
    parser = argparse.ArgumentParser(
        prog = "pypdfium2",
        description = "Command line interface to the pypdfium2 Python library",
    )
    parser.add_argument(
        "--version", "-v",
        action = "version",
        version = "pypdfium2 %s (libpdfium %s, %s)" % (
            V_PYPDFIUM2, V_LIBPDFIUM,
            "source build" if IS_SOURCEBUILD else "official build",
        ),
    )
    
    subparsers = parser.add_subparsers(dest="subcommand")
    for cmd in Subcommands.values():
        cmd.attach_parser(subparsers)
    
    if argcomplete is not None:
        argcomplete.autocomplete(parser)
    
    return parser.parse_args(argv)


def main():
    
    lib_logger = logging.getLogger("pypdfium2")
    lib_logger.addHandler(logging.StreamHandler())
    
    args = parse_args()
    if not args.subcommand:
        print("One of the following subcommands must be given: %s" % [sc for sc in Subcommands.keys()])
        return
    
    Subcommands[args.subcommand].main(args)
